import json
import uuid
from typing import List, Any

from sirius_sdk.base import JsonSerializable
from sirius_sdk.abstract.storage import AbstractImmutableCollection
from sirius_sdk.agent.wallet.abstract.non_secrets import AbstractNonSecrets, RetrieveRecordOptions


class InWalletImmutableCollection(AbstractImmutableCollection):

    DEFAULT_FETCH_LIMIT = 1000

    def __init__(self, in_wallet_storage: AbstractNonSecrets, *args, **kwargs):
        self._storage = in_wallet_storage
        self._selected_db = None
        super().__init__(*args, **kwargs)

    async def select_db(self, db_name: str):
        self._selected_db = db_name

    async def add(self, value: Any, tags: dict):
        payload = json.dumps(value)
        await self._storage.add_wallet_record(
            type_=self._selected_db,
            id_=uuid.uuid4().hex,
            value=payload,
            tags=tags
        )

    async def fetch(self, tags: dict, limit: int=None) -> (List[Any], int):
        collection, total_count = await self._storage.wallet_search(
            type_=self._selected_db,
            query=tags,
            options=RetrieveRecordOptions(retrieve_value=True),
            limit=limit or self.DEFAULT_FETCH_LIMIT
        )
        if collection:
            return [json.loads(item['value']) for item in collection], total_count
        else:
            return [], total_count



PROPOSE_NOT_ACCEPTED = "propose_not_accepted"
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = "response_processing_error"
REQUEST_NOT_ACCEPTED = "request_not_accepted"
RESPONSE_FOR_UNKNOWN_REQUEST = "response_for_unknown_request"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
VERIFY_ERROR = 'verify_error'



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Dict


class AbstractImmutableCollection(ABC):

    @abstractmethod
    async def select_db(self, db_name: str):
        raise NotImplemented

    @abstractmethod
    async def add(self, value: Any, tags: dict):
        raise NotImplemented

    @abstractmethod
    async def fetch(self, tags: dict, limit: int = None) -> (List[Any], int):
        raise NotImplemented


class AbstractKeyValueStorage(ABC):

    @abstractmethod
    async def select_db(self, db_name: str):
        raise NotImplemented

    @abstractmethod
    async def set(self, key: str, value: Any):
        raise NotImplemented

    @abstractmethod
    async def get(self, key: str) -> Optional[Any]:
        raise NotImplemented

    @abstractmethod
    async def delete(self, key: str):
        raise NotImplemented

    @abstractmethod
    async def items(self) -> Dict:
        raise NotImplemented



import json
from typing import List, Any, Optional

import sirius_sdk

from .encoding import parse_protected
from .streams import AbstractWriteOnlyStream, AbstractReadOnlyStream


class Document:

    def __init__(self):
        self.__content = None

    @property
    def content(self) -> Any:
        return self.__content

    @content.setter
    def content(self, value: Any):
        self._before_content_change()
        self.__content = value
        self._after_content_changed()

    async def save(self, stream: AbstractWriteOnlyStream):
        await stream.truncate()
        if isinstance(self.__content, str):
            await stream.write(self.__content.encode())
        else:
            await stream.write(self.__content)

    async def load(self, stream: AbstractReadOnlyStream):
        await stream.seek_to_chunk(0)
        self.__content = await stream.read()

    def _before_content_change(self):
        pass

    def _after_content_changed(self):
        pass


class EncryptedDocument(Document):

    def __init__(
            self, src: "EncryptedDocument" = None, target_verkeys: List[str] = None, content: Any = None
    ):
        super().__init__()
        self.__target_verkeys = target_verkeys or []
        self.__sender_vk = None
        self.__encrypted = False
        if src:
            self.content = src.content
            self.__encrypted = src.__encrypted
        else:
            if content is not None:
                self.content = content

    @property
    def encrypted(self) -> bool:
        return self.__encrypted

    @encrypted.setter
    def encrypted(self, value: bool):
        self.__encrypted = value

    @property
    def sender_vk(self) -> Optional[str]:
        return self.__sender_vk

    @property
    def target_verkeys(self) -> List[str]:
        return self.__target_verkeys

    @property
    def jwm(self) -> Optional[dict]:
        if self.__encrypted:
            if isinstance(self.content, bytes):
                jwm = json.loads(self.content.decode())
                return jwm
            else:
                return None
        else:
            return None

    async def encrypt(self, my_vk: str = None):
        if not self.__encrypted:
            if not self.__target_verkeys:
                raise RuntimeError(f'Target keys are missing')
            if isinstance(self.content, bytes):
                content = self.content.decode()
            else:
                content = self.content
            self.content = await sirius_sdk.Crypto.pack_message(
                message=content,
                recipient_verkeys=self.__target_verkeys,
                sender_verkey=my_vk
            )
            self.__encrypted = True

    async def decrypt(self):
        if self.__encrypted:
            protected = parse_protected(self.content)
            unpacked = await sirius_sdk.Crypto.unpack_message(self.content)
            self.content = unpacked['message']
            if isinstance(self.content, str):
                self.content = self.content.encode()
            self.__sender_vk = unpacked.get('sender_verkey')
            self.__target_verkeys = [recip['header']['kid'] for recip in protected['recipients'] if
                                     recip.get('header', {}).get('kid', None) is not None]
            self.__encrypted = False

    def _after_content_changed(self):
        self.__encrypted = False
        self.__sender_vk = None



from .messages import CoProtocolBind, CoProtocolAttach, CoProtocolInput, CoProtocolOutput, \
    CoProtocolDetach, CoProtocolProblemReport
from .state_machines import Caller, Called


__all__ = [
    "CoProtocolBind", "CoProtocolAttach", "CoProtocolInput", "CoProtocolOutput",
    "CoProtocolDetach", "CoProtocolProblemReport", "Caller", "Called"
]



import ipaddress
import json
from abc import ABCMeta, abstractmethod
from typing import Optional, Iterable

import base58
import dateutil.parser

from sirius_sdk.errors.exceptions import SiriusFieldTypeError, SiriusFieldValueError


class FieldValidator(metaclass=ABCMeta):
    """"
    Interface for field validators
    """

    optional = False

    @abstractmethod
    def validate(self, val):
        """
        Validates field value

        :param val: field value to check_for_attrs
        :return: error message or None
        """


class FieldBase(FieldValidator, metaclass=ABCMeta):
    """
    Base class for field validators
    """

    _base_types = ()

    def __init__(self, optional=False, nullable=False):
        self.optional = optional
        self.nullable = nullable

    # TODO: `check_for_attrs` should be renamed to `validation_error`
    def validate(self, val):
        """
        Performs basic validation of field value and then passes it for
        specific validation.

        :param val: field value to check_for_attrs
        :return: error message or None
        """

        if self.nullable and val is None:
            return
        type_er = self.__type_check(val)
        if type_er:
            return type_er

        spec_err = self._specific_validation(val)
        if spec_err:
            return spec_err

    @abstractmethod
    def _specific_validation(self, val):
        """
        Performs specific validation of field. Should be implemented in
        subclasses. Use it instead of overriding 'check_for_attrs'.

        :param val: field value to check_for_attrs
        :return: error message or None
        """

    def __type_check(self, val):
        if self._base_types is None:
            return  # type check is disabled
        for t in self._base_types:
            if isinstance(val, t):
                return
        return self._wrong_type_msg(val)

    def _wrong_type_msg(self, val):
        types_str = ', '.join(map(lambda x: x.__name__, self._base_types))
        return "expected types '{}', got '{}'" \
               "".format(types_str, type(val).__name__)


# TODO: The fields below should be singleton.


class AnyField(FieldBase):
    _base_types = (object,)

    def _specific_validation(self, _):
        return


class BooleanField(FieldBase):
    _base_types = (bool,)

    def _specific_validation(self, val):
        return


class IntegerField(FieldBase):
    _base_types = (int,)

    def _specific_validation(self, val):
        return


class NonEmptyStringField(FieldBase):
    _base_types = (str,)

    def _specific_validation(self, val):
        if not val:
            return 'empty string'


class LimitedLengthStringField(FieldBase):
    _base_types = (str,)

    def __init__(self, max_length: int, **kwargs):
        if not max_length > 0:
            raise SiriusFieldValueError('max_length', max_length, '> 0')
        super().__init__(**kwargs)
        self._max_length = max_length

    def _specific_validation(self, val):
        if not val:
            return 'empty string'
        if len(val) > self._max_length:
            val = val[:100] + ('...' if len(val) > 100 else '')
            return '{} is longer than {} symbols'.format(val, self._max_length)


class FixedLengthField(FieldBase):
    _base_types = (str, )

    def __init__(self, length: int, **kwargs):
        if not isinstance(length, int):
            raise TypeError('length should be integer')
        if length < 1:
            raise ValueError('should be greater than 0')
        self.length = length
        super().__init__(**kwargs)

    def _specific_validation(self, val):
        if len(val) != self.length:
            return '{} should have length {}'.format(val, self.length)


class SignatureField(LimitedLengthStringField):
    _base_types = (str, type(None))

    def _specific_validation(self, val):
        if val is None:
            # TODO do nothing because EmptySignature should be raised somehow
            return
        if len(val) == 0:
            return "signature can not be empty"
        return super()._specific_validation(val)


class RoleField(FieldBase):
    _base_types = (str, type(None))

    # TODO implement

    def _specific_validation(self, val):
        return


class NonNegativeNumberField(FieldBase):

    _base_types = (int, )

    def _specific_validation(self, val):
        if val < 0:
            return 'negative value'


class ConstantField(FieldBase):
    _base_types = None

    def __init__(self, value, **kwargs):
        super().__init__(**kwargs)
        self.value = value

    def _specific_validation(self, val):
        if val != self.value:
            return 'has to be equal {}'.format(self.value)


class IterableField(FieldBase):
    _base_types = (list, tuple)

    def __init__(self, inner_field_type: FieldValidator, min_length=None,
                 max_length=None, **kwargs):

        if not isinstance(inner_field_type, FieldValidator):
            raise SiriusFieldTypeError(
                'inner_field_type', inner_field_type, FieldValidator)

        for k in ('min_length', 'max_length'):
            m = locals()[k]
            if m is not None:
                if not isinstance(m, int):
                    raise SiriusFieldTypeError(k, m, int)
                if not m > 0:
                    raise SiriusFieldValueError(k, m, '> 0')

        self.inner_field_type = inner_field_type
        self.min_length = min_length
        self.max_length = max_length
        super().__init__(**kwargs)

    def _specific_validation(self, val):
        if self.min_length is not None:
            if len(val) < self.min_length:
                return 'length should be at least {}'.format(self.min_length)
        if self.max_length is not None:
            if len(val) > self.max_length:
                return 'length should be at most {}'.format(self.max_length)

        for v in val:
            check_er = self.inner_field_type.validate(v)
            if check_er:
                return check_er


class MapField(FieldBase):
    _base_types = (dict,)

    def __init__(self, key_field: FieldValidator,
                 value_field: FieldValidator,
                 **kwargs):
        super().__init__(**kwargs)
        self.key_field = key_field
        self.value_field = value_field

    def _specific_validation(self, val):
        for k, v in val.items():
            key_error = self.key_field.validate(k)
            if key_error:
                return key_error
            val_error = self.value_field.validate(v)
            if val_error:
                return val_error


class AnyMapField(FieldBase):
    # A map where key and value can be of arbitrary types
    _base_types = (dict,)

    def _specific_validation(self, _):
        return


class NetworkPortField(FieldBase):
    _base_types = (int,)

    def _specific_validation(self, val):
        if val <= 0 or val > 65535:
            return 'network port out of the range 0-65535'


class NetworkIpAddressField(FieldBase):
    _base_types = (str,)
    _non_valid_addresses = ('0.0.0.0', '0:0:0:0:0:0:0:0', '::')

    def _specific_validation(self, val):
        invalid_address = False
        try:
            ipaddress.ip_address(val)
        except ValueError:
            invalid_address = True
        if invalid_address or val in self._non_valid_addresses:
            return 'invalid network ip address ({})'.format(val)


class ChooseField(FieldBase):
    _base_types = None

    def __init__(self, values, **kwargs):
        self._possible_values = values
        super().__init__(**kwargs)

    def _specific_validation(self, val):
        if val not in self._possible_values:
            return "expected one of '{}', unknown value '{}'" \
                .format(', '.join(map(str, self._possible_values)), val)


class MessageField(FieldBase):
    _base_types = None

    def __init__(self, message_type, **kwargs):
        self._message_type = message_type
        super().__init__(**kwargs)

    def _specific_validation(self, val):
        if isinstance(val, self._message_type):
            return
        try:
            self._message_type(**val)
        except TypeError as ex:
            return "value {} cannot be represented as {} due to: {}" \
                .format(val, self._message_type.typename, ex)


class Base58Field(FieldBase):
    _base_types = (str,)
    _alphabet = set(base58.alphabet.decode("utf-8"))

    def __init__(self, byte_lengths: Optional[Iterable] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.byte_lengths = byte_lengths

    def _specific_validation(self, val):
        invalid_chars = set(val) - self._alphabet
        if invalid_chars:
            # only 10 chars to shorten the output
            # TODO: Why does it need to be sorted
            to_print = sorted(invalid_chars)[:10]
            return 'should not contain the following chars {}{}'.format(
                to_print, ' (truncated)' if len(to_print) < len(invalid_chars) else '')
        if self.byte_lengths is not None:
            # TODO could impact performance, need to check
            b58len = len(base58.b58decode(val))
            if b58len not in self.byte_lengths:
                return 'b58 decoded value length {} should be one of {}' \
                    .format(b58len, list(self.byte_lengths))


class FullVerkeyField(FieldBase):
    _base_types = (str,)
    _validator = Base58Field(byte_lengths=(32,))

    def _specific_validation(self, val):
        # full base58
        return self._validator.validate(val)


class AbbreviatedVerkeyField(FieldBase):
    _base_types = (str,)
    _validator = Base58Field(byte_lengths=(16,))

    def _specific_validation(self, val):
        if not val.startswith('~'):
            return 'should start with a ~'
        # abbreviated base58
        return self._validator.validate(val[1:])


# TODO: think about making it a subclass of Base58Field
class VerkeyField(FieldBase):
    _base_types = (str,)
    _b58abbreviated = AbbreviatedVerkeyField()
    _b58full = FullVerkeyField()

    def _specific_validation(self, val):
        err_ab = self._b58abbreviated.validate(val)
        err_fl = self._b58full.validate(val)
        if err_ab and err_fl:
            return 'Neither a full verkey nor an abbreviated one. One of ' \
                   'these errors should be resolved:\n {}\n{}'.\
                format(err_ab, err_fl)


class HexField(FieldBase):
    _base_types = (str,)

    def __init__(self, length=None, **kwargs):
        super().__init__(**kwargs)
        self._length = length

    def _specific_validation(self, val):
        try:
            int(val, 16)
        except ValueError:
            return "invalid hex number '{}'".format(val)
        if self._length is not None and len(val) != self._length:
            return "length should be {} length".format(self._length)


class MerkleRootField(Base58Field):
    _base_types = (str,)

    def __init__(self, *args, **kwargs):
        super().__init__(byte_lengths=(32,), *args, **kwargs)


class TimestampField(FieldBase):
    _base_types = (int,)
    _oldest_time = 1499906902

    def _specific_validation(self, val):
        if val < self._oldest_time:
            return 'should be greater than {} but was {}'. \
                format(self._oldest_time, val)


class ISODatetimeStringField(FieldBase):
    _base_types = (str,)
    parse_func = dateutil.parser.isoparse

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _specific_validation(self, val):
        try:
            self.parse_func(val)
        except Exception:
            return "{} is an invalid ISO".format(val)


class JsonField(LimitedLengthStringField):
    _base_types = (str,)

    def _specific_validation(self, val):
        # TODO: Need a mechanism to ensure a non-empty JSON if needed
        lim_str_err = super()._specific_validation(val)
        if lim_str_err:
            return lim_str_err
        try:
            json.loads(val)
        except json.decoder.JSONDecodeError:
            return 'should be a valid JSON string'


class SerializedValueField(FieldBase):
    _base_types = (bytes, str)

    def _specific_validation(self, val):
        if not val and not self.nullable:
            return 'empty serialized value'


class VersionField(LimitedLengthStringField):
    _base_types = (str,)

    def __init__(self, components_number=(3,), **kwargs):
        super().__init__(**kwargs)
        self._comp_num = components_number

    def _specific_validation(self, val):
        lim_str_err = super()._specific_validation(val)
        if lim_str_err:
            return lim_str_err
        parts = val.split(".")
        if len(parts) not in self._comp_num:
            return "version consists of {} components, but it should contain {}".format(
                len(parts), self._comp_num)
        for p in parts:
            if not p.isdigit():
                return "version component should contain only digits"
        return None


class AnyValueField(FieldBase):
    """
    Stub field validator
    """
    _base_types = None

    def _specific_validation(self, val):
        pass


class StringifiedNonNegativeNumberField(NonNegativeNumberField):
    """
    This validator is needed because of json limitations: in some cases
    numbers being converted to strings.
    """
    # TODO: Probably this should be solved another way

    _base_types = (str, int)
    _num_validator = NonNegativeNumberField()

    def _specific_validation(self, val):
        try:
            return self._num_validator.validate(int(val))
        except ValueError:
            return "stringified int expected, but was '{}'" \
                .format(val)


class DIDField(FieldBase):
    # check_for_attrs the DID to be in a proper format like did:<sov/peer>:<the did>

    _base_types = (str,)
    _valid_domains = ['sov', 'peer']

    def _specific_validation(self, val):
        did_parts = val.split(':')
        if len(did_parts) == 3 and did_parts[0] == 'did' and did_parts[1] in self._valid_domains:
            validator = Base58Field(byte_lengths=(16,))
            if not validator.validate(did_parts[2]):
                return None
        return 'Invalid DID {}'.format(val)



import asyncio


async def join(*args, timeout: int = 30):
    results = []
    items = [i for i in args]
    done, pending = await asyncio.wait(items, timeout=timeout, return_when=asyncio.FIRST_EXCEPTION)
    for f in done:
        if f.exception():
            raise f.exception()
        results.append(f.result())
    for f in pending:
        f.cancel()
    return results




TEST_MSG_TYPES = [
    'https://didcomm.org/test_protocol/1.0/request-1',
    'https://didcomm.org/test_protocol/1.0/response-1',
    'https://didcomm.org/test_protocol/1.0/request-2',
    'https://didcomm.org/test_protocol/1.0/response-2',
]
MSG_LOG = []


def check_msg_log():
    assert len(MSG_LOG) == len(TEST_MSG_TYPES)
    for i, item in enumerate(TEST_MSG_TYPES):
        assert MSG_LOG[i].type == TEST_MSG_TYPES[i]
    assert MSG_LOG[0]['content'] == 'Request1'
    assert MSG_LOG[1]['content'] == 'Response1'
    assert MSG_LOG[2]['content'] == 'Request2'
    assert MSG_LOG[3]['content'] == 'End'


def check_thread_orders():
    orders_under_checks = {}
    for i, item in enumerate(MSG_LOG):
        decorator = item.get('~thread', {}).get('received_orders', {})
        for recipient, order in decorator.items():
            if recipient in orders_under_checks:
                last_order = orders_under_checks[recipient]
                new_order = order
                assert new_order == last_order + 1
            else:
                orders_under_checks[recipient] = order



import json

from sirius_sdk.encryption import create_keypair, unpack_message, bytes_to_b58, pack_message


def test_android_device_decrypt_fault(regression_seed1: str, regression_data1: bytes):
    pk, sk = create_keypair(regression_seed1.encode())
    pk, sk = bytes_to_b58(pk), bytes_to_b58(sk)
    print('')
    message, sender_vk, recip_vk = unpack_message(
        enc_message=regression_data1,
        my_verkey=pk,
        my_sigkey=sk
    )
    js = json.loads(message)
    assert js['label'] == 'Игорь'



import asyncio
import uuid

import pytest

import sirius_sdk
from sirius_sdk import Agent
from sirius_sdk.errors.exceptions import SiriusTimeoutIO, OperationAbortedManually
from sirius_sdk.encryption.custom import bytes_to_b58
from sirius_sdk.messaging import Message, register_message_class
from tests.helpers import ServerTestSuite


class TrustPingMessageUnderTest(Message):
    pass


@pytest.mark.asyncio
async def test__all_agents_ping(test_suite: ServerTestSuite):
    for name in ['agent1', 'agent2', 'agent3', 'agent4']:
        params = test_suite.get_agent_params(name)
        agent = Agent(
            server_address=params['server_address'],
            credentials=params['credentials'],
            p2p=params['p2p'],
            timeout=5,
        )
        await agent.open()
        try:
            success = await agent.ping()
            assert success is True, 'agent [%s] is not ping-able' % name
        finally:
            await agent.close()


@pytest.mark.asyncio
async def test_agents_wallet(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    agent = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    await agent.open()
    try:
        # Check wallet calls is ok
        did, verkey = await agent.wallet.did.create_and_store_my_did()
        assert did
        assert verkey
        # check reopen is OK
        await agent.reopen()
    finally:
        await agent.close()


@pytest.mark.asyncio
async def test_subscribe_with_group_id(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    agent = Agent(
        server_address=params['server_address'],
        credentials=params['credentials'],
        p2p=params['p2p'],
        timeout=5,
    )
    await agent.open()
    try:
        await agent.subscribe(group_id='test-group')
        group_id = agent.__getattribute__('_Agent__events').balancing_group
        assert group_id == 'test-group'
    finally:
        await agent.close()


@pytest.mark.asyncio
async def test_agents_communications(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        agent2_listener = await agent2.subscribe()
        # Exchange Pairwise
        await agent1.wallet.did.store_their_did(entity2['did'], entity2['verkey'])
        if not await agent1.wallet.pairwise.is_pairwise_exists(entity2['did']):
            print('#1')
            await agent1.wallet.pairwise.create_pairwise(
                their_did=entity2['did'], my_did=entity1['did']
            )
        await agent2.wallet.did.store_their_did(entity1['did'], entity1['verkey'])
        if not await agent2.wallet.pairwise.is_pairwise_exists(entity1['did']):
            print('#2')
            await agent2.wallet.pairwise.create_pairwise(
                their_did=entity1['did'], my_did=entity2['did']
            )
        # Prepare message
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
            "comment": "Hi. Are you listening?",
            "response_requested": True
        })
        await agent1.send(
            message=trust_ping,
            their_vk=entity2['verkey'],
            endpoint=agent2_endpoint,
            my_vk=entity1['verkey'],
            routing_keys=[]
        )
        event = await agent2_listener.get_one(timeout=5)
        msg = event['message']
        assert msg['@type'] == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping'
        assert msg['@id'] == trust_ping.id
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_listener_restore_message(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        agent2_listener = await agent2.subscribe()
        # Exchange Pairwise
        await agent1.wallet.did.store_their_did(entity2['did'], entity2['verkey'])
        if not await agent1.wallet.pairwise.is_pairwise_exists(entity2['did']):
            print('#1')
            await agent1.wallet.pairwise.create_pairwise(
                their_did=entity2['did'], my_did=entity1['did']
            )
        await agent2.wallet.did.store_their_did(entity1['did'], entity1['verkey'])
        if not await agent2.wallet.pairwise.is_pairwise_exists(entity1['did']):
            print('#2')
            await agent2.wallet.pairwise.create_pairwise(
                their_did=entity1['did'], my_did=entity2['did']
            )
        # Bind Message class to protocol
        register_message_class(TrustPingMessageUnderTest, protocol='trust_ping_test')
        # Prepare message
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping_test/1.0/ping',
            "comment": "Hi. Are you listening?",
            "response_requested": True
        })
        await agent1.send(
            message=trust_ping,
            their_vk=entity2['verkey'],
            endpoint=agent2_endpoint,
            my_vk=entity1['verkey'],
            routing_keys=[]
        )
        event = await agent2_listener.get_one(timeout=5)
        msg = event['message']
        assert isinstance(msg, TrustPingMessageUnderTest), 'Unexpected msg type: ' + str(type(msg))
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_agents_trust_ping(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    entity1 = list(agent1_params['entities'].items())[0][1]
    entity2 = list(agent2_params['entities'].items())[0][1]
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        # Get endpoints
        agent1_endpoint = [e for e in agent1.endpoints if e.routing_keys == []][0].address
        agent2_endpoint = [e for e in agent2.endpoints if e.routing_keys == []][0].address
        agent1_listener = await agent1.subscribe()
        agent2_listener = await agent2.subscribe()
        # Exchange Pairwise
        await agent1.wallet.did.store_their_did(entity2['did'], entity2['verkey'])
        if not await agent1.wallet.pairwise.is_pairwise_exists(entity2['did']):
            print('#1')
            await agent1.wallet.pairwise.create_pairwise(
                their_did=entity2['did'], my_did=entity1['did']
            )
        await agent2.wallet.did.store_their_did(entity1['did'], entity1['verkey'])
        if not await agent2.wallet.pairwise.is_pairwise_exists(entity1['did']):
            print('#2')
            await agent2.wallet.pairwise.create_pairwise(
                their_did=entity1['did'], my_did=entity2['did']
            )
        # Prepare message
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping',
            "comment": "Hi. Are you listening?",
            "response_requested": True
        })
        await agent1.send(
            message=trust_ping,
            their_vk=entity2['verkey'],
            endpoint=agent2_endpoint,
            my_vk=entity1['verkey'],
            routing_keys=[]
        )
        event = await agent2_listener.get_one(timeout=5)
        assert event['recipient_verkey'] == entity2['verkey']
        assert event['sender_verkey'] == entity1['verkey']
        msg = event['message']
        assert msg['@type'] == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping'
        assert msg['@id'] == trust_ping.id

        ping_response = Message(
            {
                "@type": "did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping_response",
                "@id": "e002518b-456e-b3d5-de8e-7a86fe472847",
                "~thread": {"thid": trust_ping.id},
                "comment": "Hi yourself. I'm here."
            }
        )
        await agent2.send(
            message=ping_response,
            their_vk=entity1['verkey'],
            endpoint=agent1_endpoint,
            my_vk=entity2['verkey'],
            routing_keys=[]
        )

        event = await agent1_listener.get_one(timeout=5)
        assert event['recipient_verkey'] == entity1['verkey']
        assert event['sender_verkey'] == entity2['verkey']
        msg = event['message']
        assert msg['@type'] == 'did:sov:BzCbsNYhMrjHiqZDTUASHg;spec/trust_ping/1.0/ping_response'
        assert msg['@id'] == ping_response.id

    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_agents_crypto(test_suite: ServerTestSuite):
    agent1_params = test_suite.get_agent_params('agent1')
    agent2_params = test_suite.get_agent_params('agent2')
    agent1 = Agent(
        server_address=agent1_params['server_address'],
        credentials=agent1_params['credentials'],
        p2p=agent1_params['p2p'],
        timeout=5,
    )
    agent2 = Agent(
        server_address=agent2_params['server_address'],
        credentials=agent2_params['credentials'],
        p2p=agent2_params['p2p'],
        timeout=5,
    )
    await agent1.open()
    await agent2.open()
    try:
        did_signer = 'Th7MpTaRZVRYnPiabds81Y'
        verkey_signer = 'FYmoFw55GeQH7SRFa37dkx1d2dZ3zUF8ckg7wmL7ofN4'
        msg = b'message'
        # 1. Check sign
        signature = await agent1.wallet.crypto.crypto_sign(verkey_signer, msg)
        assert bytes_to_b58(signature) == 'QRHbNQxHLEhBuYKbe3ReTUCNRDnGDYMJvABJFEuUSFU8EzS6orRzWjMf3fR4PSgM2Z5gqfsc1kg6vYpQCCb4bjB'
        # 2. Check verify
        success = await agent2.wallet.crypto.crypto_verify(signer_vk=verkey_signer, msg=msg, signature=signature)
        assert success is True
        # 3. check verify with error
        other_msg = b'other-message'
        success = await agent2.wallet.crypto.crypto_verify(signer_vk=verkey_signer, msg=other_msg, signature=signature)
        assert success is False
    finally:
        await agent1.close()
        await agent2.close()


@pytest.mark.asyncio
async def test_bus(test_suite: ServerTestSuite):
    agent_params = test_suite.get_agent_params('agent1')
    session1 = Agent(
        server_address=agent_params['server_address'],
        credentials=agent_params['credentials'],
        p2p=agent_params['p2p'],
        timeout=5,
    )
    session2 = Agent(
        server_address=agent_params['server_address'],
        credentials=agent_params['credentials'],
        p2p=agent_params['p2p'],
        timeout=5,
    )
    await session1.open()
    await session2.open()
    try:
        thid1 = 'thread-id-' + uuid.uuid4().hex
        thid2 = 'thread-id-' + uuid.uuid4().hex
        ok = await session1.bus.subscribe(thid1)
        assert ok is True
        ok = await session1.bus.subscribe(thid2)
        assert ok is True

        content = b'Some-Message'
        sub_num = await session2.bus.publish(thid1, content)
        assert sub_num == 1

        event = await session1.bus.get_event(timeout=5)
        assert event.payload == content

        await session1.bus.unsubscribe(thid1)
        await asyncio.sleep(1)

        await session2.bus.publish(thid1, content)
        with pytest.raises(SiriusTimeoutIO):
            await session1.bus.get_event(timeout=3)

        sub_num1 = await session2.bus.publish(thid1, content)
        assert sub_num1 == 0
        sub_num2 = await session2.bus.publish(thid2, content)
        assert sub_num2 == 1
    finally:
        await session1.close()
        await session2.close()


@pytest.mark.asyncio
async def test_bus_abort(test_suite: ServerTestSuite):
    agent_params = test_suite.get_agent_params('agent1')
    session = Agent(
        server_address=agent_params['server_address'],
        credentials=agent_params['credentials'],
        p2p=agent_params['p2p'],
        timeout=5,
    )
    await session.open()
    try:
        thid = 'thread-id-' + uuid.uuid4().hex
        ok = await session.bus.subscribe(thid)
        assert ok is True

        async def __abort():
            await asyncio.sleep(1)
            await session.bus.abort()

        asyncio.ensure_future(__abort())
        with pytest.raises(OperationAbortedManually):
            await session.bus.get_event()

    finally:
        await session.close()


@pytest.mark.asyncio
async def test_bus_for_threaded_protocol(test_suite: ServerTestSuite):
    sender_params = test_suite.get_agent_params('agent1')
    receiver_params = test_suite.get_agent_params('agent2')
    sender_entity = list(sender_params['entities'].items())[0][1]
    receiver_entity = list(receiver_params['entities'].items())[0][1]
    sender = Agent(
        server_address=sender_params['server_address'],
        credentials=sender_params['credentials'],
        p2p=sender_params['p2p'],
        timeout=5,
    )
    receiver = Agent(
        server_address=receiver_params['server_address'],
        credentials=receiver_params['credentials'],
        p2p=receiver_params['p2p'],
        timeout=5,
    )
    await sender.open()
    await receiver.open()
    try:
        # Get endpoints
        receiver_endpoint = [e for e in receiver.endpoints if e.routing_keys == []][0].address
        receiver_listener = await receiver.subscribe()
        # Exchange Pairwise
        # Prepare message
        thid = 'thread-for-ping-' + uuid.uuid4().hex
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'https://didcomm.org/trust_ping/1.0/ping',
            "comment": "Hi.",
            "~thread": {"thid": thid},
        })
        # Subscribe to events
        ok = await receiver.bus.subscribe(thid)
        assert ok is True
        # Send message
        send_message_kwargs = dict(
            message=trust_ping,
            their_vk=receiver_entity['verkey'],
            endpoint=receiver_endpoint,
            my_vk=sender_entity['verkey'],
            routing_keys=[]
        )
        await sender.send(**send_message_kwargs)
        # Check income
        event = await receiver.bus.get_message(timeout=5)
        assert isinstance(event.message, Message)
        assert event.message.__class__.__name__ == 'Ping'
        assert event.message == trust_ping
        assert event.sender_verkey == sender_entity['verkey']
        assert event.recipient_verkey == receiver_entity['verkey']
        # Unsubscribe
        await receiver.bus.unsubscribe(thid)
        # send again and raise errors on reading
        await sender.send(**send_message_kwargs)
        with pytest.raises(SiriusTimeoutIO):
            await receiver.bus.get_message(timeout=3)
    finally:
        await sender.close()
        await receiver.close()


@pytest.mark.asyncio
async def test_bus_for_complex_protocol(test_suite: ServerTestSuite):
    sender_params = test_suite.get_agent_params('agent1')
    receiver_params = test_suite.get_agent_params('agent2')
    sender_entity = list(sender_params['entities'].items())[0][1]
    receiver_entity = list(receiver_params['entities'].items())[0][1]
    sender = Agent(
        server_address=sender_params['server_address'],
        credentials=sender_params['credentials'],
        p2p=sender_params['p2p'],
        timeout=5,
    )
    receiver = Agent(
        server_address=receiver_params['server_address'],
        credentials=receiver_params['credentials'],
        p2p=receiver_params['p2p'],
        timeout=5,
    )
    await sender.open()
    await receiver.open()
    try:
        # Get endpoints
        receiver_endpoint = [e for e in receiver.endpoints if e.routing_keys == []][0].address
        receiver_listener = await receiver.subscribe()
        # Exchange Pairwise
        # Prepare message
        thid = 'thread-for-ping-' + uuid.uuid4().hex
        trust_ping = Message({
            '@id': 'trust-ping-message-' + uuid.uuid4().hex,
            '@type': 'https://didcomm.org/trust_ping/1.0/ping',
            "comment": "Hi.",
        })
        some_msg = Message({
            '@id': 'some-message-' + uuid.uuid4().hex,
            '@type': 'https://didcomm.org/some-protocol/1.0/some-request',
        })
        sender_vk = sender_entity['verkey']
        receiver_vk = receiver_entity['verkey']
        # Subscribe to events
        ok, binding_ids = await receiver.bus.subscribe_ext(
            sender_vk=[sender_vk], recipient_vk=[receiver_vk], protocols=['trust_ping', 'some-protocol']
        )
        assert ok is True
        # Send message
        send_message_kwargs1 = dict(
            message=trust_ping,
            their_vk=receiver_entity['verkey'],
            endpoint=receiver_endpoint,
            my_vk=sender_entity['verkey'],
            routing_keys=[]
        )
        send_message_kwargs2 = dict(
            message=some_msg,
            their_vk=receiver_entity['verkey'],
            endpoint=receiver_endpoint,
            my_vk=sender_entity['verkey'],
            routing_keys=[]
        )
        # Send messages
        await sender.send(**send_message_kwargs1)
        await sender.send(**send_message_kwargs2)
        # Check income
        event1 = await receiver.bus.get_message(timeout=5)
        event2 = await receiver.bus.get_message(timeout=5)
        if event1.message.__class__.__name__ == 'Ping':
            assert event1.message == trust_ping
            assert event2.message == some_msg
        else:
            assert event1.message == some_msg
            assert event2.message == trust_ping
        # Unsubscribe and try again
        await receiver.bus.unsubscribe_ext(binding_ids)
        # Send messages
        await sender.send(**send_message_kwargs1)
        await sender.send(**send_message_kwargs2)
        with pytest.raises(SiriusTimeoutIO):
            await receiver.bus.get_message(timeout=5)
    finally:
        await sender.close()
        await receiver.close()


@pytest.mark.asyncio
async def test_multiple_bus_instance_for_single_connection(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    thid = 'test_multiple_bus_instance_for_single_connection-' + uuid.uuid4().hex
    content = b'Message-X'
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        bus1 = await sirius_sdk.spawn_coprotocol()
        bus2 = await sirius_sdk.spawn_coprotocol()
        await bus1.subscribe(thid)
        await bus1.publish(thid, content)
        with pytest.raises(SiriusTimeoutIO):
            await bus2.get_event(timeout=3)


@pytest.mark.asyncio
async def test_spawn_coprotocol(test_suite: ServerTestSuite):
    params = test_suite.get_agent_params('agent1')
    async with sirius_sdk.context(params['server_address'], params['credentials'], params['p2p']):
        thid = 'thid-' + uuid.uuid4().hex
        bus = await sirius_sdk.spawn_coprotocol()
        ok = await bus.subscribe(thid)
        assert ok is True

        await asyncio.sleep(1)

        payload = b'Content-Under-Test'
        num = await bus.publish(thid, payload)
        assert num > 0

        event = await bus.get_event(timeout=3)
        assert payload == event.payload



import json

import pytest

from sirius_sdk.encryption import create_keypair, pack_message, unpack_message, bytes_to_b58, sign_message, \
    verify_signed_message, did_from_verkey


@pytest.mark.asyncio
def test_sane():
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED1')
    verkey_recipient = bytes_to_b58(verkey)
    sigkey_recipient = bytes_to_b58(sigkey)
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED2')
    verkey_sender = bytes_to_b58(verkey)
    sigkey_sender = bytes_to_b58(sigkey)

    message = {
        'content': 'Test encryption строка'
    }
    message = json.dumps(message)

    packed = pack_message(
        message=message,
        to_verkeys=[verkey_recipient],
        from_verkey=verkey_sender,
        from_sigkey=sigkey_sender
    )
    unpacked, sender_vk, recip_vk = unpack_message(
        enc_message=packed,
        my_verkey=verkey_recipient,
        my_sigkey=sigkey_recipient
    )
    assert message == unpacked
    assert sender_vk, verkey_sender
    assert recip_vk, verkey_recipient


@pytest.mark.asyncio
def test_create_keypair_from_seed():
    verkey, sigkey = create_keypair(b'0000000000000000000000000000SEED')
    assert bytes_to_b58(verkey) == 'GXhjv2jGf2oT1sqMyvJtgJxNYPMHmTsdZ3c2ZYQLJExj'
    assert bytes_to_b58(sigkey) == 'xt19s1sp2UZCGhy9rNyb1FtxdKiDGZZPNFnc1KyoHNK9SDgzvPrapQPJVL9sh3e87ESLpJdwvFdxwHXagYjcaA7'


@pytest.mark.asyncio
def test_fixture():
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED1')
    verkey_recipient = bytes_to_b58(verkey)
    sigkey_recipient = bytes_to_b58(sigkey)
    verkey, sigkey = create_keypair(b'000000000000000000000000000SEED2')
    verkey_sender = bytes_to_b58(verkey)
    packed = b'{"protected": "eyJlbmMiOiAieGNoYWNoYTIwcG9seTEzMDVfaWV0ZiIsICJ0eXAiOiAiSldNLzEuMCIsICJhbGciOiAiQXV0aGNyeXB0IiwgInJlY2lwaWVudHMiOiBbeyJlbmNyeXB0ZWRfa2V5IjogInBKcW1xQS1IVWR6WTNWcFFTb2dySGx4WTgyRnc3Tl84YTFCSmtHU2VMT014VUlwT0RQWTZsMVVsaVVvOXFwS0giLCAiaGVhZGVyIjogeyJraWQiOiAiM1ZxZ2ZUcDZRNFZlRjhLWTdlVHVXRFZBWmFmRDJrVmNpb0R2NzZLR0xtZ0QiLCAic2VuZGVyIjogIjRlYzhBeFRHcWtxamd5NHlVdDF2a0poeWlYZlNUUHo1bTRKQjk1cGZSMG1JVW9KajAwWmswNmUyUEVDdUxJYmRDck8xeTM5LUhGTG5NdW5YQVJZWk5rZ2pyYV8wYTBQODJpbVdNcWNHc1FqaFd0QUhOcUw1OGNkUUYwYz0iLCAiaXYiOiAiVU1PM2o1ZHZwQnFMb2Rvd3V0c244WEMzTkVqSWJLb2oifX1dfQ==", "iv": "MchkHF2M-4hneeUJ", "ciphertext": "UgcdsV-0rIkP25eJuRSROOuqiTEXp4NToKjPMmqqtJs-Ih1b5t3EEbrrHxeSfPsHtlO6J4OqA1jc5uuD3aNssUyLug==", "tag": "sQD8qgJoTrRoyQKPeCSBlQ=="}'
    unpacked, sender_vk, recip_vk = unpack_message(
        enc_message=packed,
        my_verkey=verkey_recipient,
        my_sigkey=sigkey_recipient
    )
    message = json.dumps({
        'content': 'Test encryption строка'
    })
    assert message == unpacked
    assert sender_vk, verkey_sender
    assert recip_vk, verkey_recipient


def test_crypto_sign():
    verkey, sigkey = create_keypair(b'0000000000000000000000000000SEED')
    msg = b'message'
    signature = sign_message(message=msg, secret=sigkey)
    assert bytes_to_b58(signature) == '3tfqJYZ8ME8gTFUSHcH4uVTUx5kV7S1qPJJ65k2VtSocMfXvnzR1sbbfq6F2RcXrFtaufjEr4KQVu7aeyirYrcRm'

    success = verify_signed_message(verkey=verkey, msg=msg, signature=signature)
    assert success is True

    verkey2, sigkey2 = create_keypair(b'000000000000000000000000000SEED2')
    assert verkey2 != verkey
    signature = sign_message(message=msg, secret=sigkey2)
    success = verify_signed_message(verkey=verkey, msg=msg, signature=signature)
    assert success is False


def test_did_from_verkey():
    verkey, sigkey = create_keypair(b'0000000000000000000000000000SEED')
    assert bytes_to_b58(verkey) == 'GXhjv2jGf2oT1sqMyvJtgJxNYPMHmTsdZ3c2ZYQLJExj'
    did = did_from_verkey(verkey)
    assert bytes_to_b58(did) == 'VVZbGvuFqBdoVNY1Jh4j9Q'



from typing import Union, Optional

import sirius_sdk
from sirius_sdk.errors.exceptions import SiriusTimeoutIO


async def accept_invitation(url: str, me: sirius_sdk.Pairwise.Me, my_label: str) -> sirius_sdk.Pairwise:
    """Call this method to accept Inviter invitation that was income via URL/QR
    """
    my_endpoints = await sirius_sdk.endpoints()
    default_endpoint = [e for e in my_endpoints if not e.routing_keys][0]
    inv = sirius_sdk.aries_rfc.Invitation.from_url(url)
    proto = sirius_sdk.aries_rfc.Invitee(
        me=me,
        my_endpoint=default_endpoint
    )
    # Run protocol state-machine
    success, p2p = await proto.create_connection(
        invitation=inv,
        my_label=my_label
    )
    if success:
        return p2p
    else:
        raise RuntimeError(proto.problem_report.explain)


class InvitationManager:
    """Helper for Inviters to manage self connection-keys and P2P connections
    """

    def __init__(
            self, me: sirius_sdk.Pairwise.Me, my_label: str,
            connection_key: str = None, endpoint: sirius_sdk.Endpoint = None
    ):
        """
        :param me: My Identity
        :param my_label: My printable label
        :param connection_key: connection key to identify invitation
        :param endpoint: working end
        """
        self.__me = me
        self.__my_label = my_label
        self.__endpoint = endpoint
        self.__connection_key = connection_key

    @property
    def me(self) -> sirius_sdk.Pairwise.Me:
        return self.__me

    @property
    def my_label(self) -> str:
        return self.__my_label

    @property
    def connection_key(self) -> Optional[str]:
        return self.__connection_key

    @property
    def endpoint(self) -> Optional[sirius_sdk.Endpoint]:
        return self.__endpoint

    async def make_invitation(self) -> sirius_sdk.aries_rfc.Invitation:
        """Generate Invitation instance:
          - extract invitation-url
          - send QR to participants
        """
        if self.__connection_key is None:
            self.__connection_key = await sirius_sdk.Crypto.create_key()
        if self.__endpoint is None:
            my_endpoints = await sirius_sdk.endpoints()
            default_endpoints = [e for e in my_endpoints if not e.routing_keys]
            if default_endpoints:
                self.__endpoint = default_endpoints[0]
            else:
                self.__endpoint = my_endpoints[0]
        inv = sirius_sdk.aries_rfc.Invitation(
            label=self.__my_label, recipient_keys=[self.connection_key],
            endpoint=self.endpoint.address, routing_keys=self.__endpoint.routing_keys
        )
        return inv

    async def wait_connection(
            self, invitation: sirius_sdk.aries_rfc.Invitation, timeout: float = None
    ) -> sirius_sdk.Pairwise:
        """Wait incoming conn-requests for Invitation created earlier

        :param invitation: Invitation instance
        :param timeout: time until
        :return: P2P
        """
        co = await sirius_sdk.spawn_coprotocol()
        await co.subscribe_ext(
            sender_vk=[],
            recipient_vk=invitation.recipient_keys,
            protocols=[invitation.PROTOCOL]
        )
        try:
            while True:
                event = await co.get_message(timeout)
                if isinstance(event.message, sirius_sdk.aries_rfc.ConnRequest):
                    connection_key = event.recipient_verkey
                    proto = sirius_sdk.aries_rfc.Inviter(
                        me=self.me, connection_key=connection_key, my_endpoint=self.__endpoint
                    )
                    success, p2p = await proto.create_connection(request=event.message)
                    if success:
                        return p2p
                    else:
                        raise RuntimeError(proto.problem_report.explain)
        finally:
            await co.abort()



import json
from abc import ABC, abstractmethod
from typing import List, Union, Dict, Optional

# from sirius_sdk import acquire as resources_acquire, release as resources_release
from sirius_sdk.errors.exceptions import *

METADATA_ATTR = 'txnMetadata'
ATTR_TIME = 'txnTime'


def serialize_ordering(value: dict) -> bytes:
    data = json.dumps(value, sort_keys=True, ensure_ascii=False, separators=(',', ':')).encode()
    return data


class Transaction(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if METADATA_ATTR not in self:
            self[METADATA_ATTR] = {}

    def has_metadata(self) -> bool:
        if METADATA_ATTR in self.keys():
            meta = self[METADATA_ATTR]
            return len(meta.keys()) > 0
        else:
            return False

    @property
    def time(self) -> Optional[str]:
        return self.get(METADATA_ATTR, {}).get(ATTR_TIME)

    @time.setter
    def time(self, value: str):
        metadata = self.get(METADATA_ATTR, {})
        metadata[ATTR_TIME] = value
        self[METADATA_ATTR] = metadata

    @staticmethod
    def create(*args, **kwargs):
        inst = Transaction(*args, **kwargs)
        if inst[METADATA_ATTR] != {}:
            raise SiriusContextError('"%s" attribute must be empty for new transaction' % METADATA_ATTR)
        return inst

    @staticmethod
    def from_value(from_: Union[List[dict], dict]) -> Union[List[Dict], Dict]:
        if isinstance(from_, list):
            return [Transaction(txn) for txn in from_]
        elif isinstance(from_, dict):
            return Transaction(from_)
        else:
            raise SiriusContextError('Unexpected input value')


class MerkleInfo:

    def __init__(self, root_hash: str, audit_path: List[str]):
        self.__root_hash = root_hash
        self.__audit_path = audit_path

    @property
    def root_hash(self) -> str:
        return self.__root_hash

    @property
    def audit_path(self) -> List[str]:
        return self.__audit_path


class LedgerMeta(dict):

    def __init__(self, name: str, uid: str, created: str):
        super().__init__()
        self['name'] = name
        self['uid'] = uid
        self['created'] = created

    @property
    def name(self) -> str:
        return self['name']

    @property
    def uid(self) -> str:
        return self['uid']

    @property
    def created(self) -> str:
        return self['created']


class AuditProof(MerkleInfo):

    def __init__(self, root_hash: str, audit_path: List[str], ledger_size: int):
        super().__init__(root_hash, audit_path)
        self.__ledger_size = ledger_size

    @property
    def ledger_size(self) -> int:
        return self.__ledger_size


class AbstractMicroledger(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def size(self) -> int:
        pass

    @property
    @abstractmethod
    def uncommitted_size(self) -> int:
        pass

    @property
    @abstractmethod
    def root_hash(self) -> str:
        pass

    @property
    @abstractmethod
    def uncommitted_root_hash(self) -> str:
        pass

    @property
    @abstractmethod
    def seq_no(self) -> int:
        pass

    @abstractmethod
    async def reload(self):
        pass

    @abstractmethod
    async def rename(self, new_name: str):
        pass

    @abstractmethod
    async def init(self, genesis: List[Transaction]) -> List[Transaction]:
        pass

    @abstractmethod
    async def append(
            self, transactions: Union[List[Transaction], List[dict]], txn_time: Union[str, int] = None
    ) -> (int, int, List[Transaction]):
        pass

    @abstractmethod
    async def commit(self, count: int) -> (int, int, List[Transaction]):
        pass

    @abstractmethod
    async def discard(self, count: int):
        pass

    @abstractmethod
    async def merkle_info(self, seq_no: int) -> MerkleInfo:
        pass

    @abstractmethod
    async def audit_proof(self, seq_no: int) -> AuditProof:
        pass

    @abstractmethod
    async def reset_uncommitted(self):
        pass

    @abstractmethod
    async def get_transaction(self, seq_no: int) -> Transaction:
        pass

    @abstractmethod
    async def get_uncommitted_transaction(self, seq_no: int) -> Transaction:
        pass

    @abstractmethod
    async def get_last_transaction(self) -> Transaction:
        pass

    @abstractmethod
    async def get_last_committed_transaction(self) -> Transaction:
        pass

    @abstractmethod
    async def get_all_transactions(self) -> List[Transaction]:
        pass

    @abstractmethod
    async def get_uncommitted_transactions(self) -> List[Transaction]:
        pass


class AbstractBatchedAPI(ABC):

    @abstractmethod
    async def open(self, ledgers: Union[List[str], List[AbstractMicroledger]]) -> List[AbstractMicroledger]:
        pass

    @abstractmethod
    async def close(self):
        pass

    @abstractmethod
    async def states(self) -> List[AbstractMicroledger]:
        pass

    @abstractmethod
    async def append(
            self, transactions: Union[List[Transaction], List[dict]], txn_time: Union[str, int] = None
    ) -> List[AbstractMicroledger]:
        pass

    @abstractmethod
    async def commit(self) -> List[AbstractMicroledger]:
        pass

    @abstractmethod
    async def reset_uncommitted(self) -> List[AbstractMicroledger]:
        pass


class AbstractMicroledgerList(ABC):

    @abstractmethod
    async def create(self, name: str, genesis: Union[List[Transaction], List[dict]]) -> (AbstractMicroledger, List[Transaction]):
        pass

    @abstractmethod
    async def ledger(self, name: str) -> AbstractMicroledger:
        pass

    @abstractmethod
    async def reset(self, name: str):
        pass

    @abstractmethod
    async def is_exists(self, name: str):
        pass

    @abstractmethod
    async def leaf_hash(self, txn: Union[Transaction, bytes]) -> bytes:
        pass

    @abstractmethod
    async def list(self) -> List[LedgerMeta]:
        pass

    async def batched(self) -> Optional[AbstractBatchedAPI]:
        return None



import asyncio

import uuid
import json
import os.path
import pathlib

import pyqrcode

import sirius_sdk
from sirius_sdk.abstract.api import *
from sirius_sdk.errors.exceptions import SiriusTransportError
from sirius_sdk.messaging.transport import EndpointTransport
from sirius_sdk.messaging.forwarding import forward_wired

from .inmemory_bus import InMemoryBus


class APIDefault(APIContents, APITransport, APICoProtocols):

    DEF_TIMEOUT = 30

    def __init__(self, crypto: APICrypto = None):
        self.__transport = EndpointTransport(keepalive_timeout=self.DEF_TIMEOUT)
        self.__crypto = crypto or sirius_sdk.Crypto

    async def send(
            self, message: Message, their_vk: Union[List[str], str],
            endpoint: str, my_vk: Optional[str] = None, routing_keys: Optional[List[str]] = None
    ):
        if their_vk is None:
            their_vk = []
        if isinstance(their_vk, str):
            their_vk = [their_vk]
        if their_vk:
            payload = await self.__crypto.pack_message(
                message=json.dumps(message),
                recipient_verkeys=their_vk,
                sender_verkey=my_vk
            )
            their_vk = their_vk[0] if isinstance(their_vk, list) else their_vk
        else:
            payload = json.dumps(message).encode()
            their_vk = None

        if routing_keys:
            if my_vk:
                payload = await forward_wired(payload, their_vk, routing_keys, self.__crypto, my_vk)
            else:
                payload = await forward_wired(payload, their_vk, routing_keys)
        if routing_keys or their_vk:
            content_type = 'application/ssi-agent-wire'
        else:
            content_type = 'application/json'

        ok, body = await self.__transport.send(
            msg=payload, endpoint=endpoint, timeout=self.DEF_TIMEOUT, content_type=content_type
        )
        if not ok:
            raise SiriusTransportError(body)

    async def send_to(self, message: Message, to: Pairwise):
        await self.send(
            message=message,
            their_vk=to.their.verkey,
            endpoint=to.their.endpoint,
            my_vk=to.me.verkey,
            routing_keys=to.their.routing_keys
        )

    async def send_batched(self, message: Message, batches: List[RoutingBatch]) -> List[Tuple[bool, str]]:
        jobs = []
        for batch in batches:
            routing_keys = batch.get('routing_keys', [])
            recipient_verkeys = batch.get('recipient_verkeys', [])
            sender_verkey = batch.get('sender_verkey', None)
            endpoint_address = batch['endpoint_address']
            if routing_keys or recipient_verkeys:
                content_type = 'application/ssi-agent-wire'
            else:
                content_type = 'application/json'
            payload = await self.__crypto.pack_message(
                message=json.dumps(message),
                recipient_verkeys=recipient_verkeys,
                sender_verkey=sender_verkey
            )
            their_vk = recipient_verkeys[0] if len(recipient_verkeys) == 1 else recipient_verkeys
            if routing_keys:
                if sender_verkey:
                    payload = await forward_wired(
                        payload, their_vk, routing_keys, self.__crypto, sender_verkey
                    )
                else:
                    payload = await forward_wired(payload, their_vk, routing_keys)

            async_routine = self.__transport.send(payload, endpoint_address, self.DEF_TIMEOUT, content_type)
            jobs.append(async_routine)
        # Run simultaneously
        results = await asyncio.gather(*jobs)
        return list(results)

    async def generate_qr_code(self, value: str) -> str:
        qr = pyqrcode.create(content=value)
        base_dir = os.path.abspath(os.curdir)
        path = os.path.join(base_dir, f'qr_code_{uuid.uuid4().hex}.svg')
        qr.svg(path, scale=2)
        url = pathlib.Path(path).as_uri()
        return url

    async def spawn_coprotocol(self) -> AbstractBus:
        return InMemoryBus()



from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage
from typing import List, Union


class CoordinateMediationMessage(AriesProtocolMessage, metaclass=RegisterMessage):
    """Aries feature 0211 Message implementation

    https://github.com/hyperledger/aries-rfcs/blob/master/features/0211-route-coordination
    """

    PROTOCOL = 'coordinate-mediation'


class MediateRequest(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'mediate-request'


class MediateDeny(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'mediate-deny'


class MediateGrant(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'mediate-grant'

    def __init__(self, endpoint: str, routing_keys: List[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['endpoint'] = endpoint
        self['routing_keys'] = routing_keys


class KeylistAddAction(dict):

    def __init__(self, recipient_key: str, result: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['action'] = 'add'
        self['recipient_key'] = recipient_key
        if result is not None:
            self['result'] = result


class KeylistRemoveAction(dict):

    def __init__(self, recipient_key: str, result: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['action'] = 'remove'
        self['recipient_key'] = recipient_key
        if result is not None:
            self['result'] = result


class KeylistUpdate(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'keylist-update'

    def __init__(self, endpoint: str = None, updates: List[Union[KeylistAddAction, KeylistRemoveAction]] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if endpoint:
            self['endpoint'] = endpoint
        else:
            self['endpoint'] = None
        self['updates'] = updates or []


class KeylistUpdateResponce(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'keylist-update-responce'

    def __init__(self, updated: List[Union[KeylistAddAction, KeylistRemoveAction]], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self['updated'] = updated


class KeylistQuery(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'keylist-query'

    def __init__(self, limit: int = None, offset: int = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if limit is not None and offset is not None:
            self['paginate'] = {}
            self['paginate']['limit'] = limit
            self['paginate']['offset'] = offset


class Keylist(CoordinateMediationMessage, metaclass=RegisterMessage):

    NAME = 'keylist'

    def __init__(self, keys: List[str], count: int = None, offset: int = None, remaining: int = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self["keys"] = [{"recipient_key": key} for key in keys]
        if count is not None and offset is not None and remaining is not None:
            self['pagination'] = {}
            self['pagination']['count'] = count
            self['pagination']['offset'] = offset
            self['pagination']['remaining'] = remaining



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, List

from sirius_sdk.messaging import Message


class AbstractBus(ABC):
    """Bus concept covers problematic of wall between trust domains, typically born in combination of user-space code
    that have access to personal data
    and outdoor service that operate with low-level transport and don't have access to user-space semantic.

    Bus can help to resolve this: low-level transport services don't have access to semantic (thread-ids is not
    informative entity) on other hands user-space operations flow can run co-protocols and exchange data
    without dependency of location
    """

    @dataclass
    class BytesEvent:
        thread_id: str
        payload: bytes

    @dataclass
    class MessageEvent:
        thread_id: str
        message: Message
        sender_verkey: Optional[str]
        recipient_verkey: Optional[str]

    @abstractmethod
    async def subscribe(self, thid: str) -> bool:
        """Subscribe to events from stream marked with specific thread-id

        returns: success flag
        """
        raise NotImplemented

    @abstractmethod
    async def subscribe_ext(self, sender_vk: List[str], recipient_vk: List[str], protocols: List[str]) -> (bool, List[str]):
        """Subscribe to events that addressed by verkeys and protocol-names

         - Assumed bus agent has access to wallet secrets, so it is case fow cloud wallet agents

        returns: success flag, list of binding ids
        """
        raise NotImplemented

    @abstractmethod
    async def unsubscribe(self, thid: str):
        """UnSubscribe from events stream marked with specific thread-id
        """
        raise NotImplemented

    @abstractmethod
    async def unsubscribe_ext(self, thids: List[str]):
        """UnSubscribe from events stream marked with specific binding-ids
        """
        raise NotImplemented

    @abstractmethod
    async def publish(self, thid: str, payload: bytes) -> int:
        """Publish data to events stream marked with thread-id

        returns: num of active recipients
        """
        raise NotImplemented

    @abstractmethod
    async def get_event(self, timeout: float = None) -> BytesEvent:
        """Wait event on streams earlier subscribed

        :param timeout: wait timeout, raises SiriusTimeoutIO if timeout occurred
        returns event data in bytes
        """
        raise NotImplemented

    @abstractmethod
    async def get_message(self, timeout: float = None) -> MessageEvent:
        """Wait message on streams earlier subscribed

        :param timeout: wait timeout, raises SiriusTimeoutIO if timeout occurred
        returns parsed message
        """
        raise NotImplemented

    @abstractmethod
    async def abort(self):
        """Abort all active subscribers with raising OperationAbortedManually exception"""
        raise NotImplemented



import datetime
import json
import math
from typing import Dict

from sirius_sdk.messaging import Message, Type as MsgType
from sirius_sdk.abstract.bus import AbstractBus
from sirius_sdk.encryption import P2PConnection
from sirius_sdk.base import WebSocketConnector, INFINITE_TIMEOUT
from sirius_sdk.errors.exceptions import SiriusRPCError, SiriusTimeoutIO, OperationAbortedManually
from sirius_sdk.messaging import restore_message_instance

from .aries_rfc.feature_0753_bus.messages import *
from .aries_rfc.feature_0212_pickup.messages import PickUpNoop, PickUpProblemReport


class RpcBus(AbstractBus):

    IO_TIMEOUT = 15

    def __init__(self, connector: WebSocketConnector, p2p: P2PConnection):
        self.__connector: WebSocketConnector = connector
        self.__binding_ids: Dict[str, str] = {}
        self.__p2p = p2p
        self.__client_id = str(id(self))

    async def subscribe(self, thid: str) -> bool:
        request = BusSubscribeRequest(cast=BusSubscribeRequest.Cast(thid=thid), parent_thread_id=self.__client_id)
        resp = await self.__rfc(request)
        self.__validate(resp, expected_class=BusBindResponse)
        assert isinstance(resp, BusBindResponse)
        if resp.thread_id != thid:
            self.__set_binding_id(thid, resp.thread_id)
        return True

    async def subscribe_ext(self, sender_vk: List[str], recipient_vk: List[str], protocols: List[str]) -> (bool, List[str]):
        request = BusSubscribeRequest(
            cast=BusSubscribeRequest.Cast(
                sender_vk=sender_vk,
                recipient_vk=recipient_vk,
                protocols=protocols
            ),
            parent_thread_id=self.__client_id
        )
        resp = await self.__rfc(request)
        self.__validate(resp, expected_class=BusBindResponse)
        assert isinstance(resp, BusBindResponse)
        if isinstance(resp.thread_id, str):
            thread_ids = [resp.thread_id]
        else:
            thread_ids = resp.thread_id
        return True, thread_ids

    async def unsubscribe(self, thid: str):
        binding_id = self.__pop_binding_id(thid) or thid
        request = BusUnsubscribeRequest(thread_id=binding_id, need_answer=False)
        await self.__rfc(request, wait_response=False)

    async def unsubscribe_ext(self, thids: List[str]):
        actual_binding_ids = []
        for thid in thids:
            actual_thid = self.__pop_binding_id(thid) or thid
            actual_binding_ids.append(actual_thid)
        request = BusUnsubscribeRequest(
            thread_id=actual_binding_ids,
            need_answer=False  # don't wait response
        )
        await self.__rfc(request, wait_response=False)

    async def publish(self, thid: str, payload: bytes) -> int:
        binding_id = self.__get_binding_id(thid) or thid
        request = BusPublishRequest(thread_id=binding_id, payload=payload)
        resp = await self.__rfc(request)
        self.__validate(resp, expected_class=BusPublishResponse)
        assert isinstance(resp, BusPublishResponse)
        return resp.recipients_num

    async def get_event(self, timeout: int = None) -> AbstractBus.BytesEvent:

        if timeout is not None:
            wait_timeout = math.ceil(timeout)
            cut_stamp = datetime.datetime.now() + datetime.timedelta(seconds=wait_timeout)
        else:
            wait_timeout = INFINITE_TIMEOUT
            cut_stamp = None

        noop = PickUpNoop()
        noop.please_ack = True
        if timeout is not None:
            noop.timing = PickUpNoop.Timing(delay_milli=timeout*1000)
        await self.__connector.write(noop)

        while True:
            # Don't set timeout on websocket layer cause of message holder operate with timings on its side
            if timeout is None:
                read_timeout = None
            else:
                read_timeout = 1.1*timeout   # Increase if pickup protocol not raise response
            payload = await self.__connector.read(read_timeout)

            ok, resp = restore_message_instance(json.loads(payload.decode()))
            if ok and isinstance(resp, Message):
                thread = ThreadMixin.get_thread(resp)
                if thread and ((thread.pthid == self.__client_id) or (thread.thid == noop.id)):
                    if isinstance(resp, BusEvent):
                        return AbstractBus.BytesEvent(thread_id=resp.thread_id, payload=resp.payload)
                    elif isinstance(resp, BusBindResponse):
                        if resp.aborted is True and resp.parent_thread_id == self.__client_id:
                            raise OperationAbortedManually('Bus events awaiting was aborted by user')
                    elif isinstance(resp, BusProblemReport):
                        raise SiriusRPCError(resp.explain)
                    elif isinstance(resp, PickUpProblemReport):
                        if resp.problem_code == PickUpProblemReport.PROBLEM_CODE_TIMEOUT_OCCURRED:
                            raise SiriusTimeoutIO(resp.explain)
                        else:
                            SiriusRPCError(resp.explain)
            if wait_timeout != INFINITE_TIMEOUT:
                _ = (cut_stamp - datetime.datetime.now()).total_seconds()
                wait_timeout = math.ceil(_)
                if wait_timeout <= 0:
                    raise SiriusTimeoutIO

    async def get_message(self, timeout: float = None) -> AbstractBus.MessageEvent:
        event = await self.get_event(timeout)
        decrypted = self.__p2p.unpack(event.payload)
        if decrypted.get('@type', None):
            msg_typ = MsgType.from_str(decrypted['@type'])
            if msg_typ.protocol == 'sirius_rpc' and msg_typ.name == 'event' and 'message' in decrypted:
                ok, msg = restore_message_instance(decrypted['message'])
                if not ok:
                    msg = Message(**decrypted['message'])

                return AbstractBus.MessageEvent(
                    thread_id=event.thread_id,
                    message=msg,
                    sender_verkey=decrypted.get('sender_verkey', None),
                    recipient_verkey=decrypted.get('recipient_verkey', None)
                )
        else:
            raise SiriusRPCError('Unexpected message format')

    async def abort(self):
        request = BusUnsubscribeRequest(parent_thread_id=self.__client_id, aborted=True)
        await self.__rfc(request, wait_response=False)

    async def __rfc(self, request: BusOperation, wait_response: bool = True) -> Optional[BusOperation]:
        await self.__connector.write(request)
        if wait_response:
            payload = await self.__connector.read(timeout=self.IO_TIMEOUT)
            ok, resp = restore_message_instance(json.loads(payload.decode()))
            if ok:
                self.__validate(resp, expected_class=BusOperation)
                return resp
            else:
                raise SiriusRPCError('Unexpected response format')
        else:
            return None

    @staticmethod
    def __validate(msg: BusOperation, expected_class) -> BusOperation:
        if isinstance(msg, BusProblemReport):
            raise SiriusRPCError(msg.explain)
        elif isinstance(msg, expected_class):
            return msg
        else:
            raise SiriusRPCError(f'Unexpected response type: {msg.__class__.__name__}')

    def __get_binding_id(self, thid: str) -> Optional[str]:
        return self.__binding_ids.get(thid, None)

    def __pop_binding_id(self, thid: str) -> Optional[str]:
        bid = self.__binding_ids.get(thid, None)
        if bid is not None:
            del self.__binding_ids[thid]
        return bid

    def __set_binding_id(self, thid: str, binding_id: str):
        self.__binding_ids[thid] = binding_id



from typing import Callable, Any, Dict, List, Optional

import sirius_sdk
from sirius_sdk.hub import AbstractP2PCoProtocol
from sirius_sdk.agent.wallet.abstract.cache import CacheOptions
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.error_codes import *
from sirius_sdk.agent.aries_rfc.feature_0037_present_proof.messages import *


class SelfIdentity:
    """Wallet may contains multiple variants if attrib set to resolve proof_request
        it is tool to make decision
    """

    class SelfAttestedAttribute:
        """Self attested attribute
        """

        def __init__(self, referent_id: str, name: str, value: Any):
            self.value = value
            self.__referent_id = referent_id
            self.__name = name

        @property
        def name(self) -> str:
            return self.__name

        @property
        def referent_id(self) -> str:
            return self.__referent_id

    class CredAttribute:
        """Credential info of candidate attribute
        """
        def __init__(
                self, uid: str, cred_info: dict, revealed: bool,
                attr_name: str, on_select: Callable = None
        ):
            """
            :param cred_info: retrieved metadata from prover_search_credentials_for_proof_req wallet call
            :param revealed: make attrib revealed or No
            :param on_select(instance): callback to handle attribute was selected
            """
            self.uid = uid
            self.revealed = revealed
            self.__selected = False
            self.__cred_info = cred_info
            self.__on_select = on_select
            self.__attr_name = attr_name

        @property
        def is_selected(self) -> bool:
            """Attribute was selected in proving interaction"""
            return self.__selected

        @is_selected.setter
        def is_selected(self, value: bool):
            if self.__selected != value:
                self.__selected = value
                if self.__on_select:
                    self.__on_select(self)

        @property
        def attr_name(self) -> str:
            return self.__attr_name

        @property
        def attr_value(self) -> Any:
            return self.__cred_info.get('attrs', {}).get(self.__attr_name, None)

        @property
        def cred_id(self) -> str:
            """Credential id stored in Wallet"""
            return self.__cred_info['cred_id']

        @property
        def cred_info(self) -> dict:
            return self.__cred_info

    def __init__(self):
        # Self attested attributes as map referent_id to Attribute {referent_id: SelfAttestedAttribute}
        self.self_attested_attributes: Dict[str, SelfIdentity.SelfAttestedAttribute] = {}
        # Requested attributes as map referent_id to array of CredAttribute variants loaded from Wallet
        self.requested_attributes: Dict[str, List[SelfIdentity.CredAttribute]] = {}
        # Requested predicates as map referent_id to array of CredAttribute variants loaded from Wallet
        self.requested_predicates: Dict[str, List[SelfIdentity.CredAttribute]] = {}
        # Attributes that not found in Wallet: array of referent id
        self.non_processed = []
        #
        self.__mute = False
        self.__proof_request = None

    @property
    def proof_request(self) -> Optional[dict]:
        return self.__proof_request

    async def load(
            self, self_attested_identity: dict, proof_request: dict,
            extra_query: dict = None, limit_referents=1, default_value: Any = ''
    ) -> bool:
        self.__clear()
        self.__proof_request = proof_request
        # Stage-1: load self attested attributes
        for referent_id, data in proof_request.get('requested_attributes', {}).items():
            restrictions = data.get('restrictions', None)
            if not restrictions:
                if 'name' in data:
                    attr_name = data['name']
                    attr_value = self_attested_identity.get(attr_name, default_value)
                    self.self_attested_attributes[referent_id] = SelfIdentity.SelfAttestedAttribute(referent_id, attr_name, attr_value)
        # Stage-2: load proof-response
        proof_response = await sirius_sdk.AnonCreds.prover_search_credentials_for_proof_req(
            proof_request, extra_query=extra_query, limit_referents=limit_referents
        )
        # Stage-3: fill requested attributes
        for referent_id, cred_infos in proof_response['requested_attributes'].items():
            if len(cred_infos) > limit_referents:
                cred_infos = cred_infos[:limit_referents]  # libindy issue
            if referent_id not in self.self_attested_attributes:
                if cred_infos:
                    attr_name = proof_request['requested_attributes'].get(referent_id, {}).get('name', None)
                    attr_variants = self.requested_attributes.get(referent_id, [])
                    for i, item in enumerate(cred_infos):
                        attr_value = item.get('attrs', {}).get(attr_name, None)
                        cred_attrib = SelfIdentity.CredAttribute(
                            uid=f'{referent_id}:{i}',
                            cred_info=item['cred_info'],
                            revealed=True,
                            attr_name=attr_name,
                            on_select=self.__on_cred_attrib_select
                        )
                        if not attr_variants:
                            cred_attrib.is_selected = True
                        attr_variants.append(cred_attrib)
                    # Fill referent_id with attr alternatives variants
                    self.requested_attributes[referent_id] = attr_variants
                else:
                    self.non_processed.append(referent_id)
        # Stage-4: fill requested predicates
        for referent_id, cred_infos in proof_response['requested_predicates'].items():
            if len(cred_infos) > limit_referents:
                cred_infos = cred_infos[:limit_referents]  # libindy issue
            if cred_infos:
                attr_name = proof_request['requested_predicates'].get(referent_id, {}).get('name', None)
                attr_variants = self.requested_predicates.get(referent_id, [])
                for i, item in enumerate(cred_infos):
                    cred_attrib = SelfIdentity.CredAttribute(
                        uid=f'{referent_id}:{i}',
                        cred_info=item['cred_info'],
                        revealed=False,  # hide value for predicate requests
                        attr_name=attr_name,
                        on_select=self.__on_cred_attrib_select
                    )
                    if not attr_variants:
                        cred_attrib.is_selected = True
                    attr_variants.append(cred_attrib)
                    # Fill referent_id with attr alternatives variants
                    self.requested_predicates[referent_id] = attr_variants
                else:
                    self.non_processed.append(referent_id)
        return self.is_filled

    @property
    def is_filled(self) -> bool:
        return len(self.non_processed) == 0

    def __on_cred_attrib_select(self, emitter: CredAttribute):
        # Save from recursion calls
        if self.__mute:
            return
        # Fire
        self.__mute = True
        try:
            referent_id, index = emitter.uid.split(':')
            neighbours = self.requested_attributes.get(referent_id, []) or self.requested_predicates.get(referent_id, [])
            for neighbour in neighbours:
                if neighbour.uid != emitter.uid and neighbour.is_selected:
                    neighbour.is_selected = False
        finally:
            self.__mute = False

    def __clear(self):
        self.self_attested_attributes.clear()
        self.requested_attributes.clear()
        self.requested_predicates.clear()
        self.non_processed.clear()
        self.__proof_request = None


class ProverInteractiveMode:
    """Wrap Prover->Verifier inter-communication in single interface
    """

    def __init__(
            self, my_did: str, pool_name: str, master_secret_id: str, co: AbstractP2PCoProtocol,
            self_attested_identity: dict = None, default_value: Any = ''
    ):
        self.__self_attested_identity = self_attested_identity or {}
        self.__default_value = default_value
        self.__master_secret_id = master_secret_id
        self.__my_did = my_did
        self.__pool_name = pool_name
        self.__thread_id = None
        self.__version = '1.0'
        self.__coprotocol = co

    async def fetch(self, request: RequestPresentationMessage, extra_query: dict = None, limit_referents=1) -> SelfIdentity:
        """Fetch request correspondent data from Wallet

        :param request: Verifier request
        :param extra_query: Wallet extra-query
        :param limit_referents: max num of fetching creds
        :return: SelfIdentity
        """
        self_identity = SelfIdentity()
        await self_identity.load(
            self_attested_identity=self.__self_attested_identity,
            proof_request=request.proof_request,
            extra_query=extra_query,
            limit_referents=limit_referents,
            default_value=self.__default_value
        )
        if request.please_ack:
            self.__thread_id = request.ack_message_id
        else:
            self.__thread_id = request.id
        self.__version = request.version
        return self_identity

    async def prove(self, identity: SelfIdentity) -> (bool, Optional[PresentProofProblemReport]):
        if not identity.is_filled:
            problem_report = PresentProofProblemReport(
                REQUEST_PROCESSING_ERROR, 'No proof correspondent to proof-request'
            )
            await self.__coprotocol.send(problem_report)
            return False, problem_report
        schemas = {}
        credential_defs = {}
        rev_states = {}
        all_infos = []
        opts = CacheOptions()
        requested_credentials = {
            'self_attested_attributes': {},
            'requested_attributes': {},
            'requested_predicates': {}
        }
        # Stage-1: self attested attributes
        for referent_id, self_attest_attr in identity.self_attested_attributes.items():
            requested_credentials['self_attested_attributes'][referent_id] = self_attest_attr.value
        # Stage-2: requested attributes
        for referent_id, attr_variants in identity.requested_attributes.items():
            selected_variants = [var for var in attr_variants if var.is_selected] or [attr_variants[0]]
            selected_variant = selected_variants[0]
            info = {
                'cred_id': selected_variant.cred_info['referent'],
                'revealed': selected_variant.revealed
            }
            requested_credentials['requested_attributes'][referent_id] = info
            all_infos.append(selected_variant.cred_info)
        # Stage-3: requested predicates
        for referent_id, pred_variants in identity.requested_predicates.items():
            selected_predicates = [pred for pred in pred_variants if pred.is_selected] or [pred_variants[0]]
            selected_predicate = selected_predicates[0]
            info = {
                'cred_id': selected_predicate.cred_info['referent_id'],
            }
            if selected_predicate.revealed is True:
                info['revealed'] = True
            requested_credentials['requested_attributes'][referent_id] = info
            all_infos.append(selected_predicate.cred_info)
        # Stage-4: fill other data
        for cred_info in all_infos:
            schema_id = cred_info['schema_id']
            cred_def_id = cred_info['cred_def_id']
            schema = await sirius_sdk.Cache.get_schema(
                pool_name=self.__pool_name, submitter_did=self.__my_did,
                id_=schema_id, options=opts
            )
            cred_def = await sirius_sdk.Cache.get_cred_def(
                pool_name=self.__pool_name, submitter_did=self.__my_did,
                id_=cred_def_id, options=opts
            )
            schemas[schema_id] = schema
            credential_defs[cred_def_id] = cred_def
        # Stage-5: Build Proof
        proof = await sirius_sdk.AnonCreds.prover_create_proof(
            proof_req=identity.proof_request,
            requested_credentials=requested_credentials,
            master_secret_name=self.__master_secret_id,
            schemas=schemas,
            credential_defs=credential_defs,
            rev_states=rev_states
        )
        presentation_msg = PresentationMessage(proof, version=self.__version)
        presentation_msg.please_ack = True
        if self.__thread_id:
            presentation_msg.thread_id = self.__thread_id
        # Switch to Verifier
        ok, resp = await self.__coprotocol.switch(presentation_msg)
        if ok:
            if isinstance(resp, Ack) or isinstance(resp, PresentationAck):
                return True, None
            elif isinstance(resp, PresentProofProblemReport):
                return False, resp
            else:
                problem_report = PresentProofProblemReport(
                    RESPONSE_FOR_UNKNOWN_REQUEST, 'Unexpected response @type: %s' % str(resp.type)
                )
                await self.__coprotocol.send(problem_report)
                return False, problem_report
        else:
            problem_report = PresentProofProblemReport(
                RESPONSE_PROCESSING_ERROR, 'Response awaiting terminated by timeout'
            )
            await self.__coprotocol.send(problem_report)
            return False, problem_report



import json
import uuid
import asyncio
import datetime
from collections import OrderedDict
from dataclasses import dataclass
from typing import Union, Optional, List, OrderedDict as OrderedDictAlias

from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, VALID_DOC_URI, AriesProblemReport


class BasePickUpMessage(AriesProtocolMessage, metaclass=RegisterMessage):
    """Aries 0212 Pickup Protocol

    https://github.com/hyperledger/aries-rfcs/tree/main/features/0212-pickup

    """
    DOC_URI = VALID_DOC_URI[0]
    PROTOCOL = 'messagepickup'

    @dataclass
    class BatchedMessage:
        msg_id: str = None
        message: Union[dict, str] = None

    @property
    def return_route(self) -> Optional[str]:
        return self.get('~transport', {}).get('return_route', None)

    @return_route.setter
    def return_route(self, value: str):
        transport = self.get('~transport', {})
        transport['return_route'] = value
        self['~transport'] = transport


class PickUpStatusRequest(BasePickUpMessage):
    NAME = 'status-request'


class PickUpStatusResponse(BasePickUpMessage):
    NAME = 'status'

    def __init__(
            self,
            message_count: int = None,
            duration_limit: int = None, last_added_time: str = None,
            last_delivered_time: str = None, last_removed_time: str = None,
            *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if message_count is not None:
            self['message_count'] = message_count
        if duration_limit is not None:
            self['duration_limit'] = duration_limit
        if last_added_time is not None:
            self['last_added_time'] = last_added_time
        if last_delivered_time is not None:
            self['last_delivered_time'] = last_delivered_time
        if last_removed_time is not None:
            self['last_removed_time'] = last_removed_time

    """Required Status Properties:"""
    @property
    def message_count(self) -> Optional[int]:
        # The number of messages in the queue
        return self.get('message_count', None)

    """Optional Status Properties"""
    @property
    def duration_limit(self) -> Optional[int]:
        # The maximum duration in seconds that a message may stay in the queue
        # without being delivered (may be zero for no limit)
        return self.get('duration_limit', None)

    def last_added_time(self) -> Optional[str]:
        # A timestamp representing the last time a message was added to the queue
        return self.get('last_added_time', None)

    @property
    def last_removed_time(self) -> Optional[str]:
        # A timestamp representing the last time one or more messages was removed from the queue
        return self.get('last_removed_time', None)


class PickUpBatchRequest(BasePickUpMessage):
    NAME = 'batch-pickup'

    def __init__(self, batch_size: int = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if batch_size is not None:
            self['batch_size'] = batch_size

    @property
    def batch_size(self) -> Optional[str]:
        return self.get('batch_size', None)


class PickUpBatchResponse(BasePickUpMessage):

    NAME = 'batch'

    def __init__(self, messages: List[BasePickUpMessage.BatchedMessage] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if messages is not None:
            attach = []
            for msg in messages:
                attach.append({
                    '@id': msg.msg_id,
                    'message': msg.message
                })
            self['messages~attach'] = attach

    @property
    def messages(self) -> List[BasePickUpMessage.BatchedMessage]:
        messages = []
        for attach in self.get('messages~attach', []):
            message = BasePickUpMessage.BatchedMessage(
                msg_id=attach.get('@id', None),
                message=attach.get('message', None)
            )
            messages.append(message)
        return messages


class PickUpListRequest(BasePickUpMessage):
    NAME = 'list-pickup'

    def __init__(self, message_ids: List[str] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if message_ids is not None:
            self['message_ids'] = message_ids

    @property
    def message_ids(self) -> List[str]:
        return self.get('message_ids', [])


class PickUpListResponse(BasePickUpMessage):
    NAME = 'list-response'

    def __init__(self, messages: List[BasePickUpMessage.BatchedMessage] = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if messages is not None:
            attach = []
            for msg in messages:
                attach.append({
                    '@id': msg.msg_id,
                    'message': msg.message
                })
            self['messages~attach'] = attach

    @property
    def messages(self) -> List[BasePickUpMessage.BatchedMessage]:
        messages = []
        for attach in self.get('messages~attach', []):
            message = BasePickUpMessage.BatchedMessage(
                msg_id=attach.get('@id', None),
                message=attach.get('message', None)
            )
            messages.append(message)
        return messages


class PickUpNoop(BasePickUpMessage):
    NAME = 'noop'


class PickUpProblemReport(AriesProblemReport, metaclass=RegisterMessage):
    DOC_URI = BasePickUpMessage.DOC_URI
    PROTOCOL = BasePickUpMessage.PROTOCOL

    # Problem Codes
    PROBLEM_CODE_EMPTY = 'empty_queue'
    PROBLEM_CODE_INVALID_REQ = 'invalid_request'
    PROBLEM_CODE_TIMEOUT_OCCURRED = 'timeout_occurred'



import base64
from dataclasses import dataclass
from typing import Union, Optional, List, Any

from sirius_sdk.agent.aries_rfc.base import AriesProtocolMessage, RegisterMessage, VALID_DOC_URI, AriesProblemReport

from sirius_sdk.agent.aries_rfc.mixins import ThreadMixin


class BusOperation(AriesProtocolMessage, metaclass=RegisterMessage):
    """Aries concept 0478 Messages implementation

    hhttps://github.com/hyperledger/aries-rfcs/tree/main/concepts/0478-coprotocols
    """
    DOC_URI = VALID_DOC_URI[0]
    PROTOCOL = 'bus'

    @dataclass
    class Cast:
        thid: Union[str, List[str]] = None
        recipient_vk: Union[str, List[str]] = None
        sender_vk: Union[str, List[str]] = None
        protocols: List[str] = None

        def validate(self) -> bool:
            if self.recipient_vk or self.sender_vk:
                if not self.protocols:
                    return False
            return True

        def as_json(self) -> dict:
            js = {}
            if self.thid:
                js['thid'] = self.thid
            if self.protocols:
                js['protocols'] = sorted(self.protocols)
            if self.sender_vk:
                js['sender_vk'] = self.sender_vk if isinstance(self.sender_vk, str) else sorted(self.sender_vk)
            if self.recipient_vk:
                js['recipient_vk'] = self.recipient_vk if isinstance(self.recipient_vk, str) else sorted(self.recipient_vk)
            return js


class BusSubscribeRequest(BusOperation, metaclass=RegisterMessage):
    NAME = 'subscribe'

    def __init__(self, cast: Union[BusOperation.Cast, dict] = None, parent_thread_id: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(cast, dict):
            cast = BusOperation.Cast(**cast)
        self.__store_cast(cast)
        if parent_thread_id:
            self.thread = ThreadMixin.Thread(pthid=parent_thread_id)

    @property
    def cast(self) -> BusOperation.Cast:
        kwargs = self.get('cast', {})
        return self.Cast(**kwargs)

    @property
    def parent_thread_id(self) -> Optional[str]:
        if self.thread is not None:
            return self.thread.pthid
        else:
            return None

    def __store_cast(self, value: BusOperation.Cast = None):
        js = {}
        if value is not None:
            js = value.as_json()
        if js:
            self['cast'] = js
        elif 'cast' in self:
            del self['cast']


class BusBindResponse(BusOperation, metaclass=RegisterMessage):
    NAME = 'bind'

    def __init__(
            self, thread_id: Union[str, List[str]] = None,
            active: bool = None, parent_thread_id: str = None, aborted: bool = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        if thread_id or parent_thread_id:
            self.thread = ThreadMixin.Thread(thid=thread_id, pthid=parent_thread_id)
        if active is not None:
            self['active'] = active
        if aborted is not None:
            self['aborted'] = aborted

    @property
    def active(self) -> Optional[bool]:
        return self.get('active', None)

    @property
    def aborted(self) -> Optional[bool]:
        return self.get('aborted', None)

    @property
    def thread_id(self) -> Optional[Union[str, List[str]]]:
        thread = self.thread
        if thread:
            return thread.thid
        else:
            return None

    @property
    def parent_thread_id(self) -> Optional[str]:
        thread = self.thread
        if thread:
            return thread.pthid
        else:
            return None


class BusUnsubscribeRequest(BusBindResponse, metaclass=RegisterMessage):
    NAME = 'unsubscribe'

    def __init__(
            self, thread_id: Union[str, List[str]] = None,
            need_answer: bool = None, parent_thread_id: str = None, aborted: bool = None, *args, **kwargs
    ):
        super().__init__(thread_id=thread_id, parent_thread_id=parent_thread_id, *args, **kwargs)
        if need_answer is not None:
            self['need_answer'] = need_answer
        if aborted is not None:
            self['aborted'] = aborted

    @property
    def need_answer(self) -> Optional[bool]:
        return self.get('need_answer', None)

    @need_answer.setter
    def need_answer(self, value: bool):
        self['need_answer'] = value

    @property
    def aborted(self) -> Optional[bool]:
        return self.get('aborted', None)

    @property
    def client_id(self) -> Optional[str]:
        return self.get('client_id', None)


class BusPublishRequest(BusBindResponse, metaclass=RegisterMessage):
    NAME = 'publish'

    def __init__(self, thread_id: Union[str, List[str]] = None, payload: Any = None, *args, **kwargs):
        super().__init__(thread_id=thread_id, *args, **kwargs)
        if payload:
            self.payload = payload

    @property
    def payload(self) -> Any:
        payload = self.get('payload', {})
        if payload:
            typ = payload.get('type')
            data = payload.get('data')
            if typ == 'application/base64':
                return base64.b64decode(data.encode('ascii'))
            else:
                return data
        else:
            return None

    @payload.setter
    def payload(self, value: Any):
        if isinstance(value, dict):
            self['payload'] = value
        elif isinstance(value, bytes):
            self['payload'] = {
                'type': 'application/base64',
                'data': base64.b64encode(value).decode('ascii')
            }
        else:
            self['payload'] = {
                'type': '',
                'data': value
            }


class BusEvent(BusPublishRequest, metaclass=RegisterMessage):
    NAME = 'event'


class BusPublishResponse(BusBindResponse, metaclass=RegisterMessage):
    NAME = 'publish-result'

    def __init__(self, binding_id: Union[str, List[str]] = None, recipients_num: int = None, *args, **kwargs):
        super().__init__(binding_id, *args, **kwargs)
        if recipients_num is not None:
            self['recipients_num'] = recipients_num

    @property
    def recipients_num(self) -> Optional[int]:
        return self.get('recipients_num', None)


class BusProblemReport(AriesProblemReport, metaclass=RegisterMessage):
    DOC_URI = BusOperation.DOC_URI
    PROTOCOL = BusOperation.PROTOCOL



import logging
import contextlib

import sirius_sdk
from sirius_sdk.base import AbstractStateMachine
from sirius_sdk.errors.exceptions import SiriusValidationError, StateMachineAborted, StateMachineTerminatedWithError
from sirius_sdk.hub.coprotocols import AbstractP2PCoProtocol, CoProtocolP2PAnon
from sirius_sdk.abstract.p2p import TheirEndpoint, Pairwise
from sirius_sdk.abstract.p2p import Endpoint
from sirius_sdk.agent.aries_rfc.feature_0015_acks import Ack, Status
from sirius_sdk.agent.aries_rfc.feature_0048_trust_ping import Ping, Pong
from sirius_sdk.agent.aries_rfc.feature_0160_connection_protocol.messages import *


# Problem codes
REQUEST_NOT_ACCEPTED = "request_not_accepted"
REQUEST_PROCESSING_ERROR = 'request_processing_error'
RESPONSE_NOT_ACCEPTED = "response_not_accepted"
RESPONSE_PROCESSING_ERROR = 'response_processing_error'


class BaseConnectionStateMachine(AbstractStateMachine):

    def __init__(
            self, me: Pairwise.Me, my_endpoint: Endpoint,
            coprotocol: AbstractP2PCoProtocol = None, time_to_live: int = 60, logger=None, *args, **kwargs
    ):
        super().__init__(time_to_live=time_to_live, logger=logger, *args, **kwargs)
        self._problem_report = None
        self.__time_to_live = time_to_live
        self.__me = me
        self.__my_endpoint = my_endpoint
        self.__coprotocol = coprotocol

    @property
    def me(self) -> Pairwise.Me:
        return self.__me

    @property
    def my_endpoint(self) -> Endpoint:
        return self.__my_endpoint

    @property
    def time_to_live(self) -> Optional[int]:
        return self.__time_to_live

    @property
    def problem_report(self) -> ConnProblemReport:
        return self._problem_report

    @contextlib.asynccontextmanager
    async def coprotocol(self, endpoint: TheirEndpoint):
        co = self.__coprotocol or CoProtocolP2PAnon(
            my_verkey=self.me.verkey,
            endpoint=endpoint,
            protocols=[ConnProtocolMessage.PROTOCOL, Ack.PROTOCOL, Ping.PROTOCOL],
            time_to_live=self.time_to_live
        )
        self._register_for_aborting(co)
        try:
            try:
                yield co
            except OperationAbortedManually:
                await self.log(progress=100, message='Aborted')
                raise StateMachineAborted('Aborted by User')
        finally:
            await co.clean()
            self._unregister_for_aborting(co)


class Inviter(BaseConnectionStateMachine):
    """Implementation of Inviter role of the Aries connection protocol

    See details: https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol
    """

    def __init__(
            self, me: Pairwise.Me, connection_key: str, my_endpoint: Endpoint,
            coprotocol: AbstractP2PCoProtocol = None, time_to_live: int = 60, logger=None, *args, **kwargs
    ):
        super().__init__(
            me=me, my_endpoint=my_endpoint, coprotocol=coprotocol,
            time_to_live=time_to_live, logger=logger, *args, **kwargs
        )
        self.__connection_key = connection_key

    @property
    def connection_key(self) -> str:
        return self.__connection_key

    async def create_connection(self, request: ConnRequest, did_doc: dict = None) -> (bool, Pairwise):
        # Validate request
        await self.log(progress=0, message='Validate request', payload=dict(request), connection_key=self.connection_key)
        try:
            request.validate()
        except SiriusValidationError as e:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain=e.message,
                notify=True
            )
        else:
            await self.log(progress=20, message='Request validation OK')

        # Step 1: Extract their info from connection request
        await self.log(progress=40, message='Step-1: Extract their info from connection request')
        doc_uri = request.doc_uri
        their_did, their_vk, their_endpoint_address, their_routing_keys = request.extract_their_info()
        invitee_endpoint = TheirEndpoint(
            endpoint=their_endpoint_address,
            verkey=their_vk,
            routing_keys=their_routing_keys
        )

        # Allocate transport channel between self and theirs by verkeys factor
        async with self.coprotocol(endpoint=invitee_endpoint) as co:
            try:
                # Step 2: build connection response
                response = ConnResponse(
                    did=self.me.did,
                    verkey=self.me.verkey,
                    endpoint=self.my_endpoint.address,
                    doc_uri=doc_uri,
                    did_doc_extra=did_doc
                )
                if request.please_ack:
                    response.thread_id = request.ack_message_id
                else:
                    response.thread_id = request.id
                my_did_doc = response.did_doc
                await response.sign_connection(sirius_sdk.Crypto, self.connection_key)

                await self.log(progress=80, message='Step-2: Connection response', payload=dict(response))
                ok, ack = await co.switch(response)
                if ok:
                    if isinstance(ack, Ack) or isinstance(ack, Ping):
                        # Step 3: store their did
                        await self.log(progress=90, message='Step-3: Ack received, store their DID')
                        await sirius_sdk.DID.store_their_did(their_did, their_vk)
                        # Step 4: create pairwise
                        their = Pairwise.Their(
                            did=their_did,
                            label=request.label,
                            endpoint=their_endpoint_address,
                            verkey=their_vk,
                            routing_keys=their_routing_keys
                        )
                        their_did_doc = dict(request.did_doc)
                        metadata = {
                            'me': {
                                'did': self.me.did,
                                'verkey': self.me.verkey,
                                'did_doc': dict(my_did_doc)
                            },
                            'their': {
                                'did': their_did,
                                'verkey': their_vk,
                                'label': request.label,
                                'endpoint': {
                                    'address': their_endpoint_address,
                                    'routing_keys': their_routing_keys
                                },
                                'did_doc': their_did_doc
                            }
                        }
                        pairwise = Pairwise(me=self.me, their=their, metadata=metadata)
                        pairwise.me.did_doc = my_did_doc
                        pairwise.their.did_doc = their_did_doc
                        if isinstance(ack, Ping):
                            if ack.response_requested:
                                await co.send(Pong(ping_id=ack.id))
                        await self.log(progress=100, message='Pairwise established', payload=metadata)
                        return True, pairwise
                    elif isinstance(ack, ConnProblemReport):
                        self._problem_report = ack
                        logging.error('Code: %s; Explain: %s' % (ack.problem_code, ack.explain))
                        await self.log(
                            progress=100, message=f'Terminated with error',
                            problem_code=self._problem_report.problem_code, explain=self._problem_report.explain
                        )
                        return False, None
                    else:
                        raise StateMachineTerminatedWithError(
                            problem_code=REQUEST_PROCESSING_ERROR,
                            explain='Expect for connection response ack. Unexpected message type "%s"' % str(response.type)
                        )
                else:
                    raise StateMachineTerminatedWithError(
                        problem_code=REQUEST_PROCESSING_ERROR,
                        explain='Response ack awaiting was terminated by timeout',
                        notify=False
                    )
            except StateMachineTerminatedWithError as e:
                self._problem_report = ConnProblemReport(
                    problem_code=e.problem_code,
                    explain=e.explain,
                )
                if e.notify:
                    await co.send(self._problem_report)
                await self.log(
                    progress=100, message=f'Terminated with error',
                    problem_code=e.problem_code, explain=e.explain
                )
                return False, None


class Invitee(BaseConnectionStateMachine):
    """Implementation of Invitee role of the Aries connection protocol

    See details: https://github.com/hyperledger/aries-rfcs/tree/master/features/0160-connection-protocol
    """

    def __init__(
            self, me: Pairwise.Me, my_endpoint: Endpoint,
            coprotocol: AbstractP2PCoProtocol = None, time_to_live: int = 60, logger=None, *args, **kwargs
    ):
        super().__init__(
            me=me, my_endpoint=my_endpoint, coprotocol=coprotocol,
            time_to_live=time_to_live, logger=logger, *args, **kwargs
        )

    async def create_connection(
            self, invitation: Invitation, my_label: str, did_doc: dict = None
    ) -> (bool, Pairwise):
        # Validate invitation
        await self.log(progress=0, message='Invitation validate', payload=dict(invitation))
        try:
            invitation.validate()
        except SiriusValidationError as e:
            raise StateMachineTerminatedWithError(
                problem_code=REQUEST_PROCESSING_ERROR,
                explain='Invitation error: ' + e.message,
                notify=True
            )
        else:
            await self.log(progress=20, message='Request validation OK')
        await self.log(progress=20, message='Invitation validation OK')

        doc_uri = invitation.doc_uri
        # Extract Inviter connection_key
        connection_key = invitation.recipient_keys[0]
        inviter_endpoint = TheirEndpoint(
            endpoint=invitation.endpoint,
            verkey=connection_key,
            routing_keys=invitation.routing_keys
        )
        # Allocate transport channel between self and theirs by verkeys factor
        async with self.coprotocol(endpoint=inviter_endpoint) as co:
            await self.log(progress=40, message='Transport channel is allocated')
            try:
                request = ConnRequest(
                    label=my_label,
                    did=self.me.did,
                    verkey=self.me.verkey,
                    endpoint=self.my_endpoint.address,
                    doc_uri=doc_uri,
                    did_doc_extra=did_doc,
                    routing_keys=self.my_endpoint.routing_keys
                )

                await self.log(progress=50, message='Step-1: send connection request to Inviter', payload=dict(request))
                ok, response = await co.switch(request)
                if ok:
                    if isinstance(response, ConnResponse):
                        # Step 2: process connection response from Inviter
                        await self.log(
                            progress=40, message='Step-2: process connection response from Inviter', payload=dict(request)
                        )
                        success = await response.verify_connection(sirius_sdk.Crypto)
                        try:
                            response.validate()
                        except SiriusValidationError as e:
                            raise StateMachineTerminatedWithError(
                                problem_code=RESPONSE_NOT_ACCEPTED,
                                explain=e.message,
                                notify=True
                            )
                        if success and (response['connection~sig']['signer'] == connection_key):
                            # Step 3: extract Inviter info and store did
                            await self.log(progress=70, message='Step-3: extract Inviter info and store DID')
                            their_did, their_vk, their_endpoint_address, their_routing_keys = response.extract_their_info()
                            await sirius_sdk.DID.store_their_did(their_did, their_vk)

                            actual_endpoint = TheirEndpoint(
                                endpoint=their_endpoint_address,
                                verkey=their_vk
                            )
                            async with self.coprotocol(endpoint=actual_endpoint) as co:
                                # Step 4: Send ack to Inviter
                                if response.please_ack:
                                    ack = Ack(thread_id=response.ack_message_id, status=Status.OK)
                                    await co.send(ack)
                                    await self.log(progress=90, message='Step-4: Send ack to Inviter')
                                else:
                                    ping = Ping(comment='Connection established', response_requested=False)
                                    await co.send(ping)
                                    await self.log(progress=90, message='Step-4: Send ping to Inviter')
                                # Step 5: Make Pairwise instance
                                their = Pairwise.Their(
                                    did=their_did,
                                    label=invitation.label,
                                    endpoint=their_endpoint_address,
                                    verkey=their_vk,
                                    routing_keys=their_routing_keys
                                )
                                my_did_doc = dict(request.did_doc)
                                their_did_doc = dict(response.did_doc)
                                metadata = {
                                    'me': {
                                        'did': self.me.did,
                                        'verkey': self.me.verkey,
                                        'did_doc': my_did_doc
                                    },
                                    'their': {
                                        'did': their_did,
                                        'verkey': their_vk,
                                        'label': invitation.label,
                                        'endpoint': {
                                            'address': their_endpoint_address,
                                            'routing_keys': their_routing_keys
                                        },
                                        'did_doc': their_did_doc
                                    }
                                }
                                pairwise = Pairwise(me=self.me, their=their, metadata=metadata)
                                pairwise.me.did_doc = my_did_doc
                                pairwise.their.did_doc = their_did_doc
                                await self.log(progress=100, message='Pairwise established', payload=metadata)
                                return True, pairwise
                        else:
                            raise StateMachineTerminatedWithError(
                                problem_code=RESPONSE_NOT_ACCEPTED,
                                explain='Invalid connection response signature for connection_key: "%s"' % connection_key,
                            )
                    elif isinstance(response, ConnProblemReport):
                        self._problem_report = response
                        logging.error('Code: %s; Explain: %s' % (response.problem_code, response.explain))
                        await self.log(
                            progress=100, message=f'Terminated with error',
                            problem_code=self._problem_report.problem_code, explain=self._problem_report.explain
                        )
                        return False, None
                else:
                    raise StateMachineTerminatedWithError(
                        problem_code=RESPONSE_PROCESSING_ERROR,
                        explain='Response awaiting was terminated by timeout',
                        notify=False
                    )

            except StateMachineTerminatedWithError as e:
                self._problem_report = ConnProblemReport(
                    problem_code=e.problem_code,
                    explain=e.explain,
                )
                if e.notify:
                    await co.send(self._problem_report)
                await self.log(
                    progress=100, message=f'Terminated with error',
                    problem_code=e.problem_code, explain=e.explain
                )
                return False, None



from datetime import datetime

import pytest

import sirius_sdk
from sirius_sdk import Agent, P2PConnection
from sirius_sdk.agent.consensus.simple.state_machines import MicroLedgerSimpleConsensus
from sirius_sdk.agent.consensus.simple.messages import *

from tests.conftest import get_pairwise
from tests.helpers import run_coroutines, ServerTestSuite


async def routine_of_ledger_creator(
        uri: str, credentials: bytes, p2p: P2PConnection,
        me: Pairwise.Me, participants: List[str], ledger_name: str, genesis: List[dict]
):
    async with sirius_sdk.context(uri, credentials, p2p):
        machine = MicroLedgerSimpleConsensus(me)
        genesis = [Transaction.create(txn) for txn in genesis]
        success, ledger = await machine.init_microledger(ledger_name, participants, genesis)
        return success, ledger


async def routine_of_ledger_creation_acceptor(uri: str, credentials: bytes, p2p: P2PConnection):
    async with sirius_sdk.context(uri, credentials, p2p):
        listener = await sirius_sdk.subscribe()
        event = await listener.get_one()
        assert event.pairwise is not None
        propose = event.message
        assert isinstance(propose, InitRequestLedgerMessage)
        machine = MicroLedgerSimpleConsensus(event.pairwise.me)
        success, ledger = await machine.accept_microledger(event.pairwise, propose)
        return success, ledger


async def routine_of_txn_committer(
        uri: str, credentials: bytes, p2p: P2PConnection,
        me: Pairwise.Me, participants: List[str],
        ledger: Union[AbstractMicroledger, List[AbstractMicroledger]], txns: List[dict]
):
    async with sirius_sdk.context(uri, credentials, p2p):
        machine = MicroLedgerSimpleConsensus(me)
        txns = [Transaction.create(txn) for txn in txns]
        if isinstance(ledger, AbstractMicroledger):
            success, txns = await machine.commit(ledger, participants, txns)
        else:
            success = await machine.commit_in_parallel(ledger, participants, txns)
        return success, txns


async def routine_of_txn_acceptor(
        uri: str, credentials: bytes, p2p: P2PConnection,
        txns: List[Transaction] = None
):
    async with sirius_sdk.context(uri, credentials, p2p):
        listener = await sirius_sdk.subscribe()
        while True:
            event = await listener.get_one()
            assert event.pairwise is not None
            propose = event.message
            if isinstance(propose, ProposeTransactionsMessage):
                if txns:
                    propose['transactions'] = txns
                machine = MicroLedgerSimpleConsensus(event.pairwise.me)
                success = await machine.accept_commit(event.pairwise, propose)
                return success
            elif isinstance(propose, ProposeParallelTransactionsMessage):
                machine = MicroLedgerSimpleConsensus(event.pairwise.me)
                success = await machine.accept_commit_parallel(event.pairwise, propose)
                return success


@pytest.mark.asyncio
async def test_init_ledger_messaging(A: Agent, B: Agent, ledger_name: str):
    await A.open()
    await B.open()
    try:
        A2B = await get_pairwise(A, B)
        B2A = await get_pairwise(B, A)
        A2B.me.did = 'did:peer:' + A2B.me.did
        B2A.me.did = 'did:peer:' + B2A.me.did
        genesis_txns = [
            Transaction({"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"})
        ]
        request = InitRequestLedgerMessage(
            participants=[A2B.me.did, B2A.me.did],
            ledger_name=ledger_name,
            genesis=genesis_txns,
            root_hash='xxx'
        )

        await request.add_signature(A.wallet.crypto, A2B.me)
        await request.add_signature(B.wallet.crypto, B2A.me)

        assert len(request.signatures) == 2

        await request.check_signatures(A.wallet.crypto, A2B.me.did)
        await request.check_signatures(A.wallet.crypto, B2A.me.did)
        await request.check_signatures(A.wallet.crypto)
        await request.check_signatures(B.wallet.crypto, A2B.me.did)
        await request.check_signatures(B.wallet.crypto, B2A.me.did)
        await request.check_signatures(B.wallet.crypto)

        response = InitResponseLedgerMessage()
        response.assign_from(request)

        payload1 = dict(**request)
        payload2 = dict(**response)
        assert payload1 != payload2

        del payload1['@id']
        del payload1['@type']
        del payload2['@id']
        del payload2['@type']
        assert payload1 == payload2
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_transaction_messaging(A: Agent, B: Agent, ledger_name: str):
    await A.open()
    await B.open()
    try:
        a2b = await get_pairwise(A, B)
        b2a = await get_pairwise(B, A)
        a2b.me.did = 'did:peer:' + a2b.me.did
        b2a.me.did = 'did:peer:' + b2a.me.did
        genesis_txns = [
            Transaction({"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"})
        ]
        ledger_for_a, txns = await A.microledgers.create(ledger_name, genesis_txns)
        ledger_for_b, txns = await B.microledgers.create(ledger_name, genesis_txns)

        new_transactions = [
            Transaction({"reqId": 2, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op2"}),
            Transaction({"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"}),
        ]
        pos1, pos2, new_txns = await ledger_for_a.append(new_transactions)
        # A -> B
        state = MicroLedgerState(
                {
                    'name': ledger_for_a.name,
                    'seq_no': ledger_for_a.seq_no,
                    'size': ledger_for_a.size,
                    'uncommitted_size': ledger_for_a.uncommitted_size,
                    'root_hash': ledger_for_a.root_hash,
                    'uncommitted_root_hash': ledger_for_a.uncommitted_root_hash
                }
        )
        x = MicroLedgerState.from_ledger(ledger_for_a)
        assert state == x
        assert state.hash == x.hash
        propose = ProposeTransactionsMessage(
            transactions=new_txns,
            state=state
        )
        propose.validate()
        # B -> A
        await ledger_for_b.append(propose.transactions)
        pre_commit = PreCommitTransactionsMessage(
            state=MicroLedgerState(
                {
                    'name': ledger_for_b.name,
                    'seq_no': ledger_for_b.seq_no,
                    'size': ledger_for_b.size,
                    'uncommitted_size': ledger_for_b.uncommitted_size,
                    'root_hash': ledger_for_b.root_hash,
                    'uncommitted_root_hash': ledger_for_b.uncommitted_root_hash
                }
            )
        )
        await pre_commit.sign_state(B.wallet.crypto, b2a.me)
        pre_commit.validate()
        ok, loaded_state_hash = await pre_commit.verify_state(A.wallet.crypto, a2b.their.verkey)
        assert ok is True
        assert loaded_state_hash == state.hash
        # A -> B
        commit = CommitTransactionsMessage()
        commit.add_pre_commit(a2b.their.did, pre_commit)
        commit.validate()
        states = await commit.verify_pre_commits(A.wallet.crypto, state)
        assert a2b.their.did in str(states)
        assert a2b.their.verkey in str(states)
        # B -> A (post-commit)
        post_commit = PostCommitTransactionsMessage()
        await post_commit.add_commit_sign(B.wallet.crypto, commit, b2a.me)
        post_commit.validate()
        ok = await post_commit.verify_commits(A.wallet.crypto, commit, [a2b.their.verkey])
        assert ok is True
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_parallel_transactions_messaging(A: Agent, B: Agent, ledger_names: List[str]):
    await A.open()
    await B.open()
    try:
        txn_time = str(datetime.now())
        a2b = await get_pairwise(A, B)
        b2a = await get_pairwise(B, A)
        a2b.me.did = 'did:peer:' + a2b.me.did
        b2a.me.did = 'did:peer:' + b2a.me.did
        genesis_txns = [
            Transaction({"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"})
        ]
        ledgers_for_a = []
        ledgers_for_b = []
        for n in ledger_names:
            ledger_for_a, _ = await A.microledgers.create(n, genesis_txns)
            ledger_for_b, _ = await B.microledgers.create(n, genesis_txns)
            ledgers_for_a.append(ledger_for_a)
            ledgers_for_b.append(ledger_for_b)

        new_transactions = [
            Transaction({"reqId": 2, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op2"}),
            Transaction({"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"}),
        ]
        for txn in new_transactions:
            txn.time = txn_time
        # A -> B
        states_for_a = []
        for ledger_for_a in ledgers_for_a:
            pos1, pos2, new_txns = await ledger_for_a.append(new_transactions)
            state = MicroLedgerState.from_ledger(ledger_for_a)
            states_for_a.append(state)
        propose = ProposeParallelTransactionsMessage(transactions=new_transactions, states=states_for_a)
        propose.validate()
        # B -> A
        states_for_b = []
        for ledger_name in propose.ledgers:
            ledger_for_b = await B.microledgers.ledger(ledger_name)
            pos1, pos2, new_txns = await ledger_for_b.append(propose.transactions)
            states_for_b.append(MicroLedgerState.from_ledger(ledger_for_b))
        pre_commit = PreCommitParallelTransactionsMessage(transactions=new_transactions, states=states_for_b)
        await pre_commit.sign_states(B.wallet.crypto, b2a.me)
        pre_commit.validate()
        ok, state_hash_for_b = await pre_commit.verify_state(A.wallet.crypto, a2b.their.verkey)
        assert ok is True
        assert state_hash_for_b == propose.hash
        # A -> B
        commit = CommitParallelTransactionsMessage()
        commit.add_pre_commit(a2b.their.did, pre_commit)
        commit.validate()
        states = await commit.verify_pre_commits(A.wallet.crypto, propose.hash)
        assert a2b.their.did in str(states)
        assert a2b.their.verkey in str(states)
        # B -> A (post-commit)
        post_commit = PostCommitParallelTransactionsMessage()
        await post_commit.add_commit_sign(B.wallet.crypto, commit, b2a.me)
        post_commit.validate()
        ok = await post_commit.verify_commits(A.wallet.crypto, commit, [a2b.their.verkey])
        assert ok is True
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_parallel_batching_api_messaging(A: Agent, B: Agent, ledger_names: List[str]):
    txn_time = str(datetime.now())
    await A.open()
    await B.open()
    try:
        a2b = await get_pairwise(A, B)
        b2a = await get_pairwise(B, A)
        a2b.me.did = 'did:peer:' + a2b.me.did
        b2a.me.did = 'did:peer:' + b2a.me.did
        genesis_txns = [
            Transaction({"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"})
        ]

        ledgers_for_a = []
        ledgers_for_b = []
        for n in ledger_names:
            ledger_for_a, _ = await A.microledgers.create(n, genesis_txns)
            ledgers_for_a.append(ledger_for_a)
            ledger_for_b, _ = await B.microledgers.create(n, genesis_txns)
            ledgers_for_b.append(ledger_for_b)

        new_transactions = [
            Transaction({"reqId": 2, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op2"}),
            Transaction({"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"}),
        ]
        for txn in new_transactions:
            txn.time = txn_time

        batching_api_for_a = await A.microledgers.batched()
        batching_api_for_b = await B.microledgers.batched()
        await batching_api_for_a.open(ledger_names)
        await batching_api_for_b.open(ledger_names)
        try:
            # A -> B
            await batching_api_for_a.append(new_transactions)
            propose = ProposeParallelTransactionsMessage(
                transactions=new_transactions,
                states=[MicroLedgerState.from_ledger(item) for item in ledgers_for_a]
            )
            propose.validate()
            # B -> A
            await batching_api_for_b.append(propose.transactions)
            states_for_b = [MicroLedgerState.from_ledger(item) for item in ledgers_for_b]
            pre_commit = PreCommitParallelTransactionsMessage(states=states_for_b)
            await pre_commit.sign_states(B.wallet.crypto, b2a.me)
            pre_commit.validate()
            ok, state_hash_for_b = await pre_commit.verify_state(A.wallet.crypto, a2b.their.verkey)
            assert ok is True
            assert state_hash_for_b == propose.hash
            # A -> B
            commit = CommitParallelTransactionsMessage()
            commit.add_pre_commit(a2b.their.did, pre_commit)
            commit.validate()
            states = await commit.verify_pre_commits(A.wallet.crypto, propose.hash)
            assert a2b.their.did in str(states)
            assert a2b.their.verkey in str(states)
            # B -> A (post-commit)
            post_commit = PostCommitParallelTransactionsMessage()
            await post_commit.add_commit_sign(B.wallet.crypto, commit, b2a.me)
            post_commit.validate()
            ok = await post_commit.verify_commits(A.wallet.crypto, commit, [a2b.their.verkey])
            assert ok is True
        finally:
            await batching_api_for_a.close()
            await batching_api_for_b.close()
    finally:
        await A.close()
        await B.close()


@pytest.mark.asyncio
async def test_simple_consensus_init_ledger(
        A: Agent, B: Agent, C: Agent, ledger_name: str, test_suite: ServerTestSuite
):
    A_params = test_suite.get_agent_params('agent1')
    B_params = test_suite.get_agent_params('agent2')
    C_params = test_suite.get_agent_params('agent3')
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        coro_creator = routine_of_ledger_creator(
            A_params['server_address'], A_params['credentials'], A_params['p2p'],
            A2B.me, participants, ledger_name, genesis
        )
        coro_acceptor1 = routine_of_ledger_creation_acceptor(
            B_params['server_address'], B_params['credentials'], B_params['p2p'],
        )
        coro_acceptor2 = routine_of_ledger_creation_acceptor(
            C_params['server_address'], C_params['credentials'], C_params['p2p'],
        )

        stamp1 = datetime.now()
        print('> begin')
        await run_coroutines(coro_creator, coro_acceptor1, coro_acceptor2, timeout=30)
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')

        is_exists_for_A = await A.microledgers.is_exists(ledger_name)
        is_exists_for_B = await B.microledgers.is_exists(ledger_name)
        is_exists_for_C = await C.microledgers.is_exists(ledger_name)
        assert is_exists_for_A
        assert is_exists_for_B
        assert is_exists_for_C

        for agent in [A, B, C]:
            ledger = await agent.microledgers.ledger(ledger_name)
            txns = await ledger.get_all_transactions()
            assert len(txns) == 2
            assert 'op1' in str(txns)
            assert 'op2' in str(txns)
    finally:
        await A.close()
        await B.close()
        await C.close()


@pytest.mark.asyncio
async def test_simple_consensus_commit(
        A: Agent, B: Agent, C: Agent, ledger_name: str, test_suite: ServerTestSuite
):
    A_params = test_suite.get_agent_params('agent1')
    B_params = test_suite.get_agent_params('agent2')
    C_params = test_suite.get_agent_params('agent3')
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        ledger_for_a, _ = await A.microledgers.create(ledger_name, genesis)
        await B.microledgers.create(ledger_name, genesis)
        await C.microledgers.create(ledger_name, genesis)

        txns = [
            {"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5"},
        ]

        coro_committer = routine_of_txn_committer(
            A_params['server_address'], A_params['credentials'], A_params['p2p'],
            A2B.me, participants, ledger_for_a, txns
        )
        coro_acceptor1 = routine_of_txn_acceptor(
            B_params['server_address'], B_params['credentials'], B_params['p2p'],
        )
        coro_acceptor2 = routine_of_txn_acceptor(
            C_params['server_address'], C_params['credentials'], C_params['p2p'],
        )

        stamp1 = datetime.now()
        print('> begin')
        await run_coroutines(coro_committer, coro_acceptor1, coro_acceptor2, timeout=60)
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')

        ledger_for_a = await A.microledgers.ledger(ledger_name)
        ledger_for_b = await B.microledgers.ledger(ledger_name)
        ledger_for_c = await C.microledgers.ledger(ledger_name)
        for ledger in [ledger_for_a, ledger_for_b, ledger_for_c]:
            all_txns = await ledger.get_all_transactions()
            assert 'op3' in str(all_txns)
            assert 'op4' in str(all_txns)
            assert 'op5' in str(all_txns)
    finally:
        await A.close()
        await B.close()
        await C.close()


@pytest.mark.asyncio
async def test_simple_consensus_commit_parallel(
        A: Agent, B: Agent, C: Agent, ledger_names: List[str], test_suite: ServerTestSuite
):
    A_params = test_suite.get_agent_params('agent1')
    B_params = test_suite.get_agent_params('agent2')
    C_params = test_suite.get_agent_params('agent3')
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        txns = [
            {"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5"},
        ]
        # Init ledgers
        leader_ledgers = []
        for name in ledger_names:
            ledger_of_leader, _ = await A.microledgers.create(name, genesis)
            leader_ledgers.append(ledger_of_leader)
            await B.microledgers.create(name, genesis)
            await C.microledgers.create(name, genesis)

        coro_committer = routine_of_txn_committer(
            A_params['server_address'], A_params['credentials'], A_params['p2p'],
            A2B.me, participants, leader_ledgers, txns
        )
        coro_acceptor1 = routine_of_txn_acceptor(
            B_params['server_address'], B_params['credentials'], B_params['p2p'],
        )
        coro_acceptor2 = routine_of_txn_acceptor(
            C_params['server_address'], C_params['credentials'], C_params['p2p'],
        )

        stamp1 = datetime.now()
        print('> begin')
        await run_coroutines(
            coro_committer,
            coro_acceptor1,
            coro_acceptor2,
            timeout=60
        )
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')

        for name in ledger_names:
            ledger_for_a = await A.microledgers.ledger(name)
            ledger_for_b = await B.microledgers.ledger(name)
            ledger_for_c = await C.microledgers.ledger(name)
            for ledger in [ledger_for_a, ledger_for_b, ledger_for_c]:
                all_txns = await ledger.get_all_transactions()
                assert 'op3' in str(all_txns)
                assert 'op4' in str(all_txns)
                assert 'op5' in str(all_txns)
    finally:
        await A.close()
        await B.close()
        await C.close()


@pytest.mark.asyncio
async def test_simple_consensus_error(A: Agent, B: Agent, C: Agent, ledger_name: str, test_suite: ServerTestSuite):
    A_params = test_suite.get_agent_params('agent1')
    B_params = test_suite.get_agent_params('agent2')
    C_params = test_suite.get_agent_params('agent3')
    await A.open()
    await B.open()
    await C.open()
    try:
        A2B = await get_pairwise(A, B)
        A2C = await get_pairwise(A, C)
        assert A2B.me == A2C.me
        B2A = await get_pairwise(B, A)
        B2C = await get_pairwise(B, C)
        assert B2A.me == B2C.me
        C2A = await get_pairwise(C, A)
        C2B = await get_pairwise(C, B)
        assert C2A.me == C2B.me
        participants = [
            A2B.me.did,
            A2B.their.did,
            A2C.their.did
        ]
        genesis = [
            {"reqId": 1, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op1"},
            {"reqId": 2, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op2"}
        ]
        ledger_for_a, _ = await A.microledgers.create(ledger_name, genesis)
        initial_state_root_hash = ledger_for_a.root_hash
        await B.microledgers.create(ledger_name, genesis)
        await C.microledgers.create(ledger_name, genesis)

        txns = [
            {"reqId": 3, "identifier": "5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3"},
            {"reqId": 4, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4"},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5"},
        ]
        broken_txns = [
            {"reqId": 3, "identifier": "BROKEN-5rArie7XKukPCaEwq5XGQJnM9Fc5aZE3M9HAPVfMU2xC", "op": "op3", "txnMetadata": {"seqNo": 4}},
            {"reqId": 4, "identifier": "BROKEN-2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op4", "txnMetadata": {"seqNo": 5}},
            {"reqId": 5, "identifier": "2btLJAAb1S3x6hZYdVyAePjqtQYi2ZBSRGy4569RZu8h", "op": "op5", "txnMetadata": {"seqNo": 6}},
        ]
        broken_txns = [Transaction(txn) for txn in broken_txns]
        coro_committer = routine_of_txn_committer(
            A_params['server_address'], A_params['credentials'], A_params['p2p'],
            A2B.me, participants, ledger_for_a, txns
        )
        coro_acceptor1 = routine_of_txn_acceptor(
            B_params['server_address'], B_params['credentials'], B_params['p2p'], broken_txns
        )
        coro_acceptor2 = routine_of_txn_acceptor(
            C_params['server_address'], C_params['credentials'], C_params['p2p'], broken_txns
        )

        stamp1 = datetime.now()
        print('> begin')
        results = await run_coroutines(coro_committer, coro_acceptor1, coro_acceptor2, timeout=60)
        print('> end')
        stamp2 = datetime.now()
        delta = stamp2 - stamp1
        print(f'***** Consensus timeout: {delta.seconds}')
        for res in results:
            if type(res) is tuple:
                val = res[0]
            else:
                val = res
            assert val is False

        ledger_for_a = await A.microledgers.ledger(ledger_name)
        ledger_for_b = await B.microledgers.ledger(ledger_name)
        ledger_for_c = await C.microledgers.ledger(ledger_name)
        for ledger in [ledger_for_a, ledger_for_b, ledger_for_c]:
            await ledger.reload()
            assert ledger.root_hash == initial_state_root_hash
            all_txns = await ledger.get_all_transactions()
            assert 'op3' not in str(all_txns)
            assert 'op4' not in str(all_txns)
            assert 'op5' not in str(all_txns)
    finally:
        await A.close()
        await B.close()
        await C.close()



from sirius_sdk.agent.aries_rfc.feature_0211_mediator_coordination_protocol.messages import *


__all__ = [
    "MediateRequest", "MediateDeny", "MediateGrant", "KeylistAddAction", "KeylistRemoveAction",
    "KeylistUpdate", "KeylistUpdateResponce", "KeylistQuery", "Keylist"
]



from typing import List, Optional
from urllib.parse import urlparse, urlunparse


class Endpoint:
    """Active Agent endpoints
    https://github.com/hyperledger/aries-rfcs/tree/master/concepts/0094-cross-domain-messaging
    """

    def __init__(self, address: str, routing_keys: List[str], is_default: bool=False):
        self.__url = address
        self.__routing_keys = routing_keys
        self.__is_default = is_default

    @property
    def address(self):
        return self.__url

    @property
    def routing_keys(self) -> List[str]:
        return self.__routing_keys

    @property
    def is_default(self):
        return self.__is_default


class TheirEndpoint:

    def __init__(self, endpoint: str, verkey: str, routing_keys: List[str]=None):
        self.endpoint = endpoint
        self.verkey = verkey
        self.routing_keys = routing_keys or []

    @property
    def address(self) -> str:
        return self.endpoint

    @property
    def netloc(self) -> Optional[str]:
        if self.endpoint:
            return urlparse(self.endpoint).netloc
        else:
            return None

    @netloc.setter
    def netloc(self, value: str):
        if self.endpoint:
            components = list(urlparse(self.endpoint))
            components[1] = value
            self.endpoint = urlunparse(components)


class Pairwise:

    class Their(TheirEndpoint):

        def __init__(
                self, did: str, label: str, endpoint: str, verkey: str,
                routing_keys: List[str] = None, did_doc: dict = None
        ):
            self.did = did
            self.label = label
            self.did_doc = did_doc
            super().__init__(endpoint, verkey, routing_keys)

    class Me:

        def __init__(self, did, verkey, did_doc: dict = None):
            self.did = did
            self.verkey = verkey
            self.did_doc = did_doc

        def __eq__(self, other):
            if isinstance(other, Pairwise.Me):
                return self.did == other.did and self.verkey == other.verkey and self.did_doc == other.did_doc

    def __init__(self, me: Me, their: Their, metadata: dict=None):
        self.__me = me
        self.__their = their
        self.__metadata = metadata

    @property
    def their(self) -> Their:
        return self.__their

    @property
    def me(self) -> Me:
        return self.__me

    @property
    def metadata(self) -> dict:
        return self.__metadata